# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

# pylint: disable=C0103

"""
Curve fitting widgets
=====================

Overview
--------

The :mod:`.widgets.fit` module provides interactive curve fitting widgets allowing:

* to fit data manually (by moving sliders)
* or automatically (with standard optimization algorithms provided by `scipy`).

The :func:`guifit` function is a factory function that returns a dialog box
allowing to fit data with a given function.

Example
-------

Here is an example of use of the :func:`guifit` function:

.. literalinclude:: ../../plotpy/tests/features/test_fit.py
   :start-after: guitest:


.. image:: /images/screenshots/fit.png

Reference
---------

.. autofunction:: guifit

.. autoclass:: FitDialog
   :members:
.. autoclass:: FitParam
   :members:
.. autoclass:: AutoFitParam
   :members:
"""

from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import guidata
import numpy as np
from guidata.configtools import get_icon
from guidata.dataset import (
    BoolItem,
    ChoiceItem,
    DataSet,
    FloatItem,
    IntItem,
    StringItem,
    restore_dataset,
    update_dataset,
)
from guidata.qthelpers import create_groupbox, exec_dialog

# Do not remove this import (used by optimization funcs)
from numpy import inf  # noqa: F401
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW
from qtpy.QtWidgets import QWidget  # only to help intersphinx find QWidget
from scipy.optimize import fmin, fmin_bfgs, fmin_cg, fmin_l_bfgs_b, fmin_powell, leastsq

from plotpy.builder import make
from plotpy.config import _
from plotpy.plot import PlotDialog, PlotOptions, PlotWidget

if TYPE_CHECKING:
    from plotpy.items import XRangeSelection
    from plotpy.panels import PanelWidget


class AutoFitParam(DataSet):
    """Automatic fit parameters"""

    xmin = FloatItem("xmin", default=0.0)
    xmax = FloatItem("xmax", default=1.0)
    method = ChoiceItem(
        _("Method"),
        [
            ("simplex", "Simplex"),
            ("powel", "Powel"),
            ("bfgs", "BFGS"),
            ("l_bfgs_b", "L-BFGS-B"),
            ("cg", _("Conjugate Gradient")),
            ("lq", _("Least squares")),
        ],
        default="lq",
    )
    err_norm = StringItem(
        "enorm",
        default="2.0",
        help=_("for simplex, powel, cg and bfgs norm used by the error function"),
    )
    xtol = FloatItem(
        "xtol", default=0.0001, help=_("for simplex, powel, least squares")
    )
    ftol = FloatItem(
        "ftol", default=0.0001, help=_("for simplex, powel, least squares")
    )
    gtol = FloatItem("gtol", default=0.0001, help=_("for cg, bfgs"))
    norm = StringItem(
        "norm", default="inf", help=_("for cg, bfgs. inf is max, -inf is min")
    )


class FitParamDataSet(DataSet):
    """Fit parameter dataset"""

    name = StringItem(_("Name"), default="")
    value = FloatItem(_("Value"), default=0.0)
    min = FloatItem(_("Min"), default=-1.0)
    max = FloatItem(_("Max"), default=1.0).set_pos(col=1)
    steps = IntItem(_("Steps"), default=5000)
    format = StringItem(_("Format"), default="%.3f").set_pos(col=1)
    logscale = BoolItem(_("Logarithmic"), _("Scale"))
    unit = StringItem(_("Unit"), default="").set_pos(col=1)
    locked = BoolItem(_("Fixed value during optimization"), _("Lock"))


class FitParam:
    """Fit parameters

    Args:
        name: name of the parameter
        value: value of the parameter
        min: minimum value of the parameter
        max: maximum value of the parameter
        logscale: if True, the parameter is fitted in logscale. Default is False.
        steps: number of steps for the slider. Default is 5000.
        format: format of the parameter. Default is "%.3f".
        size_offset: size offset of the parameter. Default is 0.
        unit: unit of the parameter. Default is "".
        locked: if True, the parameter value is locked and will not be modified
            by the automatic fit. Default is False.
    """

    def __init__(
        self,
        name: str,
        value: float,
        min: float,
        max: float,
        logscale: bool = False,
        steps: int = 5000,
        format: str = "%.3f",
        size_offset: int = 0,
        unit: str = "",
        locked: bool = False,
    ):
        self.name = name
        self.value = value
        self.min = min
        self.max = max
        self.logscale = logscale
        self.steps = steps
        self.format = format
        self.unit = unit
        self.locked = locked
        self.prefix_label = None
        self.lineedit = None
        self.unit_label = None
        self.slider = None
        self.button = None
        self._widgets = []
        self._size_offset = size_offset
        self._refresh_callback = None
        self.dataset = FitParamDataSet(title=_("Curve fitting parameter"))

    def copy(self) -> FitParam:
        """Return a copy of this fitparam

        Returns:
            Copy of this fitparam
        """
        return self.__class__(
            self.name,
            self.value,
            self.min,
            self.max,
            self.logscale,
            self.steps,
            self.format,
            self._size_offset,
            self.unit,
            self.locked,
        )

    def create_widgets(self, parent: QWidget, refresh_callback: Callable) -> None:
        """Create widgets

        Args:
            parent: parent widget
            refresh_callback: callback function to refresh the plot
        """
        self._refresh_callback = refresh_callback
        self.prefix_label = QW.QLabel()
        font = self.prefix_label.font()
        font.setPointSize(font.pointSize() + self._size_offset)
        self.prefix_label.setFont(font)
        self.button = QW.QPushButton()
        self.button.setIcon(get_icon("settings.png"))
        self.button.setToolTip(
            _("Edit '{name}' fit parameter properties").format(name=self.name)
        )
        self.button.clicked.connect(lambda: self.edit_param(parent))
        self.lineedit = QW.QLineEdit()
        self.lineedit.editingFinished.connect(self.line_editing_finished)
        self.unit_label = QW.QLabel(self.unit)
        self.slider = QW.QSlider()
        self.slider.setOrientation(QC.Qt.Horizontal)
        self.slider.setRange(0, self.steps - 1)
        self.slider.valueChanged.connect(self.slider_value_changed)
        self.update(refresh=False)
        self.add_widgets(
            [
                self.prefix_label,
                self.lineedit,
                self.unit_label,
                self.slider,
                self.button,
            ]
        )

    def add_widgets(self, widgets: list[QWidget]) -> None:
        """Add widgets

        Args:
            widgets: list of widgets to add
        """
        self._widgets += widgets

    def get_widgets(self) -> list[QWidget]:
        """Get widgets

        Returns:
            List of widgets
        """
        return self._widgets

    def set_scale(self, state: int) -> None:
        """Set scale

        Args:
            state: state
        """
        self.logscale = state > 0
        self.update_slider_value()

    def set_text(self, fmt: str = None) -> None:
        """Set text

        Args:
            fmt: format (default: None)
        """
        style = "<span style='color: #444444'><b>{}</b></span>"
        if self.locked:
            style = "<span style='color: #888888'><b>{} 🔒</b></span>"
        self.prefix_label.setText(style.format(self.name))
        if self.value is None:
            value_str = ""
        else:
            if fmt is None:
                fmt = self.format
            value_str = fmt % self.value
        self.lineedit.setText(value_str)
        is_disabled = bool(
            self.locked or (self.value == self.min and self.max == self.min)
        )
        self.lineedit.setDisabled(is_disabled)

    def line_editing_finished(self):
        """Line editing finished"""
        try:
            self.value = float(self.lineedit.text())
        except ValueError:
            self.set_text()
        self.update_slider_value()
        self._refresh_callback()

    def slider_value_changed(self, int_value: int) -> None:
        """Slider value changed

        Args:
            int_value: integer value
        """
        if self.logscale:
            total_delta = np.log10(1 + self.max - self.min)
            self.value = (
                self.min + 10 ** (total_delta * int_value / (self.steps - 1)) - 1
            )
        else:
            total_delta = self.max - self.min
            self.value = self.min + total_delta * int_value / (self.steps - 1)
        self.set_text()
        self._refresh_callback()

    def update_slider_value(self):
        """Update slider value"""
        if self.value is None or self.min is None or self.max is None:
            self.slider.setEnabled(False)
            if self.slider.parentWidget() and self.slider.parentWidget().isVisible():
                self.slider.show()
        elif self.value == self.min and self.max == self.min:
            self.slider.hide()
        else:
            self.slider.setEnabled(not self.locked)
            if self.slider.parentWidget() and self.slider.parentWidget().isVisible():
                self.slider.show()
            if self.logscale:
                value_delta = max([np.log10(1 + self.value - self.min), 0.0])
                total_delta = np.log10(1 + self.max - self.min)
            else:
                value_delta = self.value - self.min
                total_delta = self.max - self.min
            intval = int(self.steps * value_delta / total_delta)
            self.slider.blockSignals(True)
            self.slider.setValue(intval)
            self.slider.blockSignals(False)

    def edit_param(self, parent: QWidget) -> None:
        """Edit param

        Args:
            parent: parent widget
        """
        update_dataset(self.dataset, self)
        if self.dataset.edit(parent=parent):
            restore_dataset(self.dataset, self)
            if self.value > self.max:
                self.max = self.value
            if self.value < self.min:
                self.min = self.value
            self.update()

    def update(self, refresh: bool = True) -> None:
        """Update

        Args:
            refresh: refresh (default: True)
        """
        self.unit_label.setText(self.unit)
        self.slider.setRange(0, self.steps - 1)
        self.update_slider_value()
        self.set_text()
        if refresh:
            self._refresh_callback()


def add_fitparam_widgets_to(
    layout: QW.QGridLayout,
    fitparams: list[FitParam],
    refresh_callback: Callable,
    param_cols: int = 1,
    stretch_col: int = 1,
) -> None:
    """Add fitparam widgets to layout

    Args:
        layout: layout
        fitparams: list of fitparams
        refresh_callback: refresh callback
        param_cols: number of columns (default: 1)
        stretch_col: stretch column (default: 1)
    """
    row_contents = []
    row_nb = 0
    col_nb = 0
    for i, param in enumerate(fitparams):
        param.create_widgets(layout.parentWidget(), refresh_callback)
        widgets = param.get_widgets()
        w_colums = len(widgets) + 1
        row_contents += [
            (widget, row_nb, j + col_nb * w_colums) for j, widget in enumerate(widgets)
        ]
        col_nb += 1
        if col_nb == param_cols:
            row_nb += 1
            col_nb = 0
    for widget, row, col in row_contents:
        layout.addWidget(widget, row, col)
    if fitparams:
        for col_nb in range(param_cols):
            layout.setColumnStretch(stretch_col + col_nb * w_colums, 5)
            if col_nb > 0:
                layout.setColumnStretch(col_nb * w_colums - 1, 1)


class FitWidget(QWidget):
    """Fit widget

    Args:
        parent: parent widget (default: None)
        param_cols: number of columns (default: 1)
        legend_anchor: legend anchor (default: "TR")
        auto_fit: auto fit (default: False)
    """

    SIG_TOGGLE_VALID_STATE = QC.Signal(bool)

    def __init__(
        self,
        parent: QWidget = None,
        param_cols: int = 1,
        legend_anchor: str = "TR",
        auto_fit: bool = False,
    ) -> None:
        super().__init__(parent)
        self.x = None
        self.y = None
        self.fitfunc = None
        self.fitargs = None
        self.fitkwargs = None
        self.fitparams = None
        self.autofit_prm = None

        self.data_curve = None
        self.fit_curve = None
        self.legend = None
        self.legend_anchor = legend_anchor
        self.xrange = None
        self.show_xrange = False

        self.param_cols = param_cols
        self.auto_fit_enabled = auto_fit
        self.button_list: list[QW.QPushButton] = []

        self.params_layout: QW.QGridLayout = None
        self.plot_widget: PlotWidget = None

        self.setup_widget()

    def set_plot_widget(self, plot_widget: PlotWidget) -> None:
        """Set plot widget

        Args:
            plot_widget: plot widget
        """
        self.plot_widget = plot_widget
        plot_widget.plot.SIG_RANGE_CHANGED.connect(self.range_changed)
        self.refresh()

    def resizeEvent(self, event) -> None:
        """Reimplement Qt method

        Args:
            event: event
        """
        super().resizeEvent(event)
        if self.plot_widget is not None:
            self.plot_widget.plot.replot()

    def setup_widget(self) -> None:
        """Setup widget"""
        fit_layout = QW.QHBoxLayout()
        self.params_layout = QW.QGridLayout()
        params_group = create_groupbox(
            self, _("Fit parameters"), layout=self.params_layout
        )
        if self.auto_fit_enabled:
            auto_group = self.create_autofit_group()
            fit_layout.addWidget(auto_group)
        fit_layout.addWidget(params_group)
        self.setLayout(fit_layout)

    def create_autofit_group(self) -> QW.QGroupBox:
        """Create autofit group

        Returns:
            Autofit group
        """
        auto_button = QW.QPushButton(get_icon("apply.png"), _("Run"), self)
        auto_button.clicked.connect(self.autofit)
        autoprm_button = QW.QPushButton(get_icon("settings.png"), _("Settings"), self)
        autoprm_button.clicked.connect(self.edit_parameters)
        xrange_button = QW.QPushButton(get_icon("xrange.png"), _("Bounds"), self)
        xrange_button.setCheckable(True)
        xrange_button.toggled.connect(self.toggle_xrange)
        auto_layout = QW.QVBoxLayout()
        auto_layout.addWidget(auto_button)
        auto_layout.addWidget(autoprm_button)
        auto_layout.addWidget(xrange_button)
        self.button_list += [auto_button, autoprm_button, xrange_button]
        return create_groupbox(self, _("Automatic fit"), layout=auto_layout)

    # Public API ---------------------------------------------------------------
    def set_data(
        self,
        x: np.ndarray,
        y: np.ndarray,
        fitfunc: Callable = None,
        fitparams: list[FitParam] | None = None,
        fitargs: tuple | None = None,
        fitkwargs: dict | None = None,
    ) -> None:
        """Set fit data

        Args:
            x: x data
            y: y data
            fitfunc: fit function. Defaults to None.
            fitparams: fit parameters. Defaults to None.
            fitargs: fit args. Defaults to None.
            fitkwargs: fit kwargs. Defaults to None.
        """
        if self.fitparams is not None and fitparams is not None:
            self.clear_params_layout()
        self.x = x
        self.y = y
        if fitfunc is not None:
            self.fitfunc = fitfunc
        if fitparams is not None:
            self.fitparams = fitparams
        if fitargs is not None:
            self.fitargs = fitargs
        if fitkwargs is not None:
            self.fitkwargs = fitkwargs
        self.autofit_prm = AutoFitParam(title=_("Automatic fitting options"))
        self.autofit_prm.xmin = x.min()
        self.autofit_prm.xmax = x.max()
        self.compute_imin_imax()
        if self.fitparams is not None and fitparams is not None:
            self.populate_params_layout()
        self.refresh()

    def set_fit_data(
        self,
        fitfunc: Callable,
        fitparams: list[FitParam],
        fitargs: tuple | None = None,
        fitkwargs: dict | None = None,
    ) -> None:
        """Set fit data

        Args:
            fitfunc: fit function
            fitparams: fit parameters
            fitargs: fit args. Defaults to None.
            fitkwargs: fit kwargs. Defaults to None.
        """
        if self.fitparams is not None:
            self.clear_params_layout()
        self.fitfunc = fitfunc
        self.fitparams = fitparams
        self.fitargs = fitargs
        self.fitkwargs = fitkwargs
        self.populate_params_layout()
        self.refresh()

    def clear_params_layout(self) -> None:
        """Clear params layout"""
        for i, param in enumerate(self.fitparams):
            for widget in param.get_widgets():
                if widget is not None:
                    self.params_layout.removeWidget(widget)
                    widget.hide()

    def populate_params_layout(self) -> None:
        """Populate params layout"""
        add_fitparam_widgets_to(
            self.params_layout, self.fitparams, self.refresh, param_cols=self.param_cols
        )

    def get_fitfunc_arguments(self) -> tuple[list, dict]:
        """Return fitargs and fitkwargs

        Returns:
            Fitargs and fitkwargs
        """
        fitargs = self.fitargs
        if self.fitargs is None:
            fitargs = []
        fitkwargs = self.fitkwargs
        if self.fitkwargs is None:
            fitkwargs = {}
        return fitargs, fitkwargs

    def refresh(self) -> None:
        """Refresh Fit Tool dialog box"""
        # Update button states
        enable = (
            self.x is not None
            and self.y is not None
            and self.x.size > 0
            and self.y.size > 0
            and self.fitfunc is not None
            and self.fitparams is not None
            and len(self.fitparams) > 0
        )
        for btn in self.button_list:
            btn.setEnabled(enable)
        self.SIG_TOGGLE_VALID_STATE.emit(enable)

        if not enable:
            # Fit widget is not yet configured
            return

        fitargs, fitkwargs = self.get_fitfunc_arguments()
        yfit = self.fitfunc(
            self.x, [p.value for p in self.fitparams], *fitargs, **fitkwargs
        )

        plot = self.plot_widget.plot

        if self.legend is None:
            self.legend = make.legend(anchor=self.legend_anchor)
            plot.add_item(self.legend)

        if self.xrange is None:
            self.xrange = make.xrange(0.0, 1.0)
            plot.add_item(self.xrange)
        self.xrange.set_range(self.autofit_prm.xmin, self.autofit_prm.xmax)
        self.xrange.setVisible(self.show_xrange)

        if self.data_curve is None:
            self.data_curve = make.curve([], [], _("Data"), color="b", linewidth=2.0)
            plot.add_item(self.data_curve)
        self.data_curve.set_data(self.x, self.y)

        if self.fit_curve is None:
            self.fit_curve = make.curve([], [], _("Fit"), color="r", linewidth=2.0)
            plot.add_item(self.fit_curve)
        self.fit_curve.set_data(self.x, yfit)

        plot.replot()
        plot.disable_autoscale()

    def range_changed(
        self, xrange_obj: XRangeSelection, xmin: float, xmax: float
    ) -> None:  # pylint: disable=unused-argument
        """Range changed

        Args:
            xrange_obj: xrange object
            xmin: xmin
            xmax: xmax
        """
        self.autofit_prm.xmin, self.autofit_prm.xmax = xmin, xmax
        self.compute_imin_imax()

    def toggle_xrange(self, state: bool) -> None:
        """Toggle xrange visibility

        Args:
            state: state
        """
        self.xrange.setVisible(state)
        self.plot_widget.plot.replot()
        if state:
            self.plot_widget.plot.set_active_item(self.xrange)
        else:
            #  If the button is unckeked then set to the complete range
            self.autofit_prm.xmin = self.x.min()
            self.autofit_prm.xmax = self.x.max()
            self.xrange.set_range(self.autofit_prm.xmin, self.autofit_prm.xmax)
        self.show_xrange = state

    def edit_parameters(self) -> None:
        """Edit fit parameters"""
        if self.autofit_prm.edit(parent=self):
            self.xrange.set_range(self.autofit_prm.xmin, self.autofit_prm.xmax)
            self.plot_widget.plot.replot()
            self.compute_imin_imax()

    def compute_imin_imax(self) -> None:
        """Compute i_min and i_max"""
        self.i_min = self.x.searchsorted(self.autofit_prm.xmin)
        self.i_max = self.x.searchsorted(self.autofit_prm.xmax, side="right")

    def get_full_params(self, free_params: np.ndarray) -> list[float]:
        """Build full parameter list from free parameters

        Args:
            free_params: values of unlocked parameters only

        Returns:
            Full parameter list with locked parameters at their fixed values
        """
        full_params = []
        free_idx = 0
        for p in self.fitparams:
            if p.locked:
                full_params.append(p.value)
            else:
                full_params.append(free_params[free_idx])
                free_idx += 1
        return full_params

    def errorfunc(self, free_params: np.ndarray) -> np.ndarray:
        """Get error function

        Args:
            free_params: values of unlocked fit parameters

        Returns:
            Error function
        """
        x = self.x[self.i_min : self.i_max]
        y = self.y[self.i_min : self.i_max]
        fitargs, fitkwargs = self.get_fitfunc_arguments()
        params = self.get_full_params(free_params)
        return y - self.fitfunc(x, params, *fitargs, **fitkwargs)

    def autofit(self) -> None:
        """Autofit"""
        meth = self.autofit_prm.method

        # Extract only unlocked parameters for optimization
        free_params = np.array([p.value for p in self.fitparams if not p.locked])

        # If all parameters are locked, nothing to optimize
        if len(free_params) == 0:
            return

        if meth == "lq":
            x = self.autofit_lq(free_params)
        elif meth == "simplex":
            x = self.autofit_simplex(free_params)
        elif meth == "powel":
            x = self.autofit_powel(free_params)
        elif meth == "bfgs":
            x = self.autofit_bfgs(free_params)
        elif meth == "l_bfgs_b":
            x = self.autofit_l_bfgs(free_params)
        elif meth == "cg":
            x = self.autofit_cg(free_params)
        else:
            return

        # Restore optimized values only to unlocked parameters
        free_idx = 0
        for p in self.fitparams:
            if not p.locked:
                p.value = x[free_idx]
                free_idx += 1

        self.refresh()
        for prm in self.fitparams:
            prm.update()

    def get_norm_func(self) -> Callable:
        """Get norm function

        Returns:
            Norm function
        """
        prm = self.autofit_prm
        err_norm = eval(prm.err_norm)

        def func(params):
            """

            :param params:
            :return:
            """
            err = np.linalg.norm(self.errorfunc(params), err_norm)
            return err

        return func

    def autofit_simplex(self, x0: np.ndarray) -> np.ndarray:
        """Autofit using simplex

        Args:
            x0: initial value

        Returns:
            Fitted values
        """
        prm = self.autofit_prm

        x = fmin(self.get_norm_func(), x0, xtol=prm.xtol, ftol=prm.ftol)
        return x

    def autofit_powel(self, x0: np.ndarray) -> np.ndarray:
        """Autofit using Powell

        Args:
            x0: initial value

        Returns:
            Fitted values
        """
        prm = self.autofit_prm

        x = fmin_powell(self.get_norm_func(), x0, xtol=prm.xtol, ftol=prm.ftol)
        return x

    def autofit_bfgs(self, x0: np.ndarray) -> np.ndarray:
        """Autofit using BFGS

        Args:
            x0: initial value

        Returns:
            Fitted values
        """
        prm = self.autofit_prm

        x = fmin_bfgs(self.get_norm_func(), x0, gtol=prm.gtol, norm=eval(prm.norm))
        return x

    def autofit_l_bfgs(self, x0: np.ndarray) -> np.ndarray:
        """Autofit using L-BFGS-B

        Args:
            x0: initial value

        Returns:
            Fitted values
        """
        prm = self.autofit_prm
        # Build bounds only for unlocked parameters
        bounds = [(p.min, p.max) for p in self.fitparams if not p.locked]

        x, _f, _d = fmin_l_bfgs_b(
            self.get_norm_func(), x0, pgtol=prm.gtol, approx_grad=1, bounds=bounds
        )
        return x

    def autofit_cg(self, x0: np.ndarray) -> np.ndarray:
        """Autofit using conjugate gradient

        Args:
            x0: initial value

        Returns:
            Fitted values
        """
        prm = self.autofit_prm

        x = fmin_cg(self.get_norm_func(), x0, gtol=prm.gtol, norm=eval(prm.norm))
        return x

    def autofit_lq(self, x0: np.ndarray) -> np.ndarray:
        """Autofit using leastsq

        Args:
            x0: initial value

        Returns:
            Fitted values
        """
        prm = self.autofit_prm

        def func(params: list[float]) -> np.ndarray:
            """Error function

            Args:
                params: fit parameter values

            Returns:
                Error function
            """
            err = self.errorfunc(params)
            return err

        x, _ier = leastsq(func, x0, xtol=prm.xtol, ftol=prm.ftol)
        return x

    def get_values(self) -> list[float]:
        """Convenience method to get fit parameter values

        Returns:
            Fit parameter values
        """
        return [param.value for param in self.fitparams]


class FitDialog(PlotDialog):
    """Fit dialog box

    Args:
        title: window title. Defaults to None.
        icon: window icon. Defaults to "plotpy.svg".
        edit: enable/disable edit menu. Defaults to True.
        toolbar: enable/disable toolbar. Defaults to False.
        auto_tools: enable/disable auto tools. Defaults to True.
        options: plot options. Defaults to None.
        parent: parent widget. Defaults to None.
        panels: list of panel widgets. Defaults to None.
        param_cols: number of columns for parameter table.
         Defaults to 1.
        legend_anchor: legend anchor. Defaults to "TR".
        auto_fit: enable/disable auto fit. Defaults to False.
    """

    def __init__(
        self,
        title: str | None = None,
        icon: str = "plotpy.svg",
        edit: bool = True,
        toolbar: bool = False,
        auto_tools: bool = True,
        options: PlotOptions | dict[str, Any] | None = None,
        parent: QWidget | None = None,
        panels: list[PanelWidget] | None = None,
        param_cols: int = 1,
        legend_anchor: str = "TR",
        auto_fit: bool = False,
    ):
        super().__init__(
            title=title if title is not None else _("Curve fitting"),
            icon=icon,
            edit=edit,
            toolbar=toolbar,
            auto_tools=auto_tools,
            options=options,
            parent=parent,
            panels=panels,
        )
        self.fit_widget = fitw = FitWidget(
            self,
            param_cols=param_cols,
            legend_anchor=legend_anchor,
            auto_fit=auto_fit,
        )
        fitw.set_plot_widget(self.plot_widget)
        self.add_widget(self.fit_widget)
        ok_btn = self.button_box.button(QW.QDialogButtonBox.Ok)
        self.fit_widget.SIG_TOGGLE_VALID_STATE.connect(ok_btn.setEnabled)
        self.setWindowFlags(QC.Qt.Window)

    def set_data(
        self,
        x: np.ndarray,
        y: np.ndarray,
        fitfunc: Callable = None,
        fitparams: list[FitParam] | None = None,
        fitargs: tuple | None = None,
        fitkwargs: dict | None = None,
    ) -> None:
        """Set fit data

        Args:
            x: x data
            y: y data
            fitfunc: fit function. Defaults to None.
            fitparams: fit parameters. Defaults to None.
            fitargs: fit args. Defaults to None.
            fitkwargs: fit kwargs. Defaults to None.
        """
        self.fit_widget.set_data(x, y, fitfunc, fitparams, fitargs, fitkwargs)

    def get_values(self) -> list[float]:
        """Returns fit parameter values

        Returns:
            Fit parameter values
        """
        return self.fit_widget.get_values()

    def autofit(self) -> None:
        """Autofit"""
        self.fit_widget.autofit()


def guifit(
    x: np.ndarray,
    y: np.ndarray,
    fitfunc: Callable,
    fitparams: list[FitParam] | None = None,
    fitargs: tuple | None = None,
    fitkwargs: dict | None = None,
    wintitle: str | None = None,
    title: str | None = None,
    xlabel: str | None = None,
    ylabel: str | None = None,
    param_cols: int = 1,
    auto_fit: bool = True,
    winsize: tuple[int, int] | None = None,
    winpos: tuple[int, int] | None = None,
) -> list[float] | None:
    """GUI-based curve fitting tool

    Args:
        x: x data
        y: y data
        fitfunc: fit function
        fitparams: fit parameters. Defaults to None.
        fitargs: fit args. Defaults to None.
        fitkwargs: fit kwargs. Defaults to None.
        wintitle: window title. Defaults to None.
        title: plot title. Defaults to None.
        xlabel: x label. Defaults to None.
        ylabel: y label. Defaults to None.
        param_cols: number of columns for fit parameters. Defaults to 1.
        auto_fit: auto fit. Defaults to True.
        winsize (tuple[int, int] | None): window size. Defaults to None.
        winpos (tuple[int, int] | None): window position. Defaults to None.

    Returns:
        Fit parameter values or None if the user cancels the dialog
    """
    _app = guidata.qapplication()
    win = FitDialog(
        edit=True,
        title=wintitle,
        toolbar=True,
        param_cols=param_cols,
        auto_fit=auto_fit,
        options=PlotOptions(title=title, xlabel=xlabel, ylabel=ylabel, type="curve"),
    )
    win.set_data(x, y, fitfunc, fitparams, fitargs, fitkwargs)
    if winsize is not None:
        win.resize(*winsize)
    if winpos is not None:
        win.move(*winpos)
    if exec_dialog(win):
        return win.get_values()


if __name__ == "__main__":
    x = np.linspace(-10, 10, 1000)
    y = np.cos(1.5 * x) + np.random.rand(x.shape[0]) * 0.2

    def fit(x: np.ndarray, params: list[float]) -> np.ndarray:
        """
        Fit function

        Args:
            x: x data
            params: fit parameter values

        Returns:
            Fit values
        """
        a, b = params
        return np.cos(b * x) + a

    a = FitParam("Offset", 1.0, 0.0, 2.0)
    b = FitParam("Frequency", 1.05, 0.0, 10.0, logscale=True)
    params = [a, b]
    values = guifit(x, y, fit, params, auto_fit=True)
    print(values)
    print([param.value for param in params])



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import qtpy.QtCore as QC
from guidata.qthelpers import exec_dialog, qt_app_context

from plotpy.coords import canvas_to_axes
from plotpy.interfaces.items import ICurveItemType
from plotpy.tests.unit.utils import (
    CLICK,
    create_window,
    drag_mouse,
    keyboard_event,
    mouse_event_at_relative_plot_pos,
)
from plotpy.tools import EditPointTool, SelectPointsTool, SelectPointTool
from plotpy.tools.curve import DownSamplingTool

if TYPE_CHECKING:
    from plotpy.items.curve.base import CurveItem


from plotpy.items import DataInfoLabel, XRangeSelection
from plotpy.tools import CurveStatsTool
from plotpy.tools.selection import SelectTool


def test_curve_stat_tool():
    """Test the curve stats tool."""
    with qt_app_context(exec_loop=False):
        win, _tool = create_window(CurveStatsTool)
        win.show()
        plot = win.manager.get_plot()

        original_stat_items = [
            item
            for item in plot.get_items()
            if isinstance(item, (DataInfoLabel, XRangeSelection))
        ]

        drag_mouse(win, np.array([0.4, 0.6]), np.array([0.5, 0.5]))

        new_stat_items = [
            item
            for item in plot.get_items()
            if isinstance(item, (DataInfoLabel, XRangeSelection))
        ]

        assert len(new_stat_items) == len(original_stat_items) + 2

        # There should be one more new DataInfoLabel and one new XRangeSelection
        # compared to the original plot items (before mouse drag)

        win.manager.add_tool(SelectTool).activate()
        plot.unselect_all()

        mouse_event_at_relative_plot_pos(win, (0.5, 0.5), CLICK)
        selected_items = plot.get_selected_items()
        assert len(selected_items) == 1 and isinstance(
            selected_items[0], XRangeSelection
        )
        range_item: XRangeSelection = selected_items[0]

        x00, x01, y0 = range_item.get_handles_pos()
        drag_mouse(win, np.linspace(0.5, 0.9, 100), np.full(100, 0.5))
        x10, x11, y1 = range_item.get_handles_pos()

        assert x00 < x10 and x01 < x11 and y0 == y1

        curve_item: CurveItem = plot.get_items(item_type=ICurveItemType)[0]  # type: ignore
        current_pos = canvas_to_axes(curve_item, QC.QPointF(x10, y1))
        x_min, _ = plot.get_axis_limits(curve_item.xAxis())

        assert current_pos is not None
        range_item.move_shape(current_pos, (x_min, 0))
        assert np.isclose(range_item.get_range()[0], x_min)

        exec_dialog(win)


def test_free_select_point_tool():
    """Test the free select point tool."""
    with qt_app_context(exec_loop=False):
        win, tool = create_window(SelectPointTool)
        win.show()
        mouse_event_at_relative_plot_pos(win, (0.5, 0.5), CLICK)
        assert tool.get_coordinates() is not None
        exec_dialog(win)


def test_contrained_select_point_tool():
    """Test the constrained select point tool contrained to a CurveItem."""
    with qt_app_context(exec_loop=False):
        win, tool = create_window(SelectPointTool)
        win.show()
        tool.on_active_item = True

        mouse_event_at_relative_plot_pos(win, (0.5, 0.5), CLICK)
        coor = tool.get_coordinates()
        curve_item: CurveItem = win.manager.get_plot().get_active_item()  # type: ignore
        arr_x, arr_y = curve_item.get_data()
        assert coor is not None

        x, y = coor
        assert np.where(arr_x == x)[0] == np.where(arr_y == y)[0]
        exec_dialog(win)


def test_select_points_tool():
    """Test the select points tool constrained to a CurveItem."""
    with qt_app_context(exec_loop=False):
        win, tool = create_window(tool_class=SelectPointsTool)
        mod = QC.Qt.KeyboardModifier.ControlModifier

        mouse_event_at_relative_plot_pos(win, (0.4, 0.5), CLICK, mod)
        assert len(tool.get_coordinates() or ()) == 1

        mouse_event_at_relative_plot_pos(win, (0.5, 0.5), CLICK, mod)
        mouse_event_at_relative_plot_pos(win, (0.8, 0.8), CLICK, mod)
        print(tool.get_coordinates())
        assert len(tool.get_coordinates() or ()) == 3

        mouse_event_at_relative_plot_pos(win, (0.8, 0.8), CLICK, mod)
        assert len(tool.get_coordinates() or ()) == 2

        mouse_event_at_relative_plot_pos(win, (0.7, 0.5), CLICK, mod)
        assert len(tool.get_coordinates() or ()) == 3

        mouse_event_at_relative_plot_pos(win, (0.1, 0.1), CLICK)
        assert len(tool.get_coordinates() or ()) == 1

        coor = tool.get_coordinates()
        assert coor is not None
        x, y = coor[0]
        curve_item: CurveItem = win.manager.get_plot().get_active_item()  # type: ignore
        arr_x, arr_y = curve_item.get_data()
        assert np.where(arr_x == x)[0] == np.where(arr_y == y)[0]

        exec_dialog(win)


def test_edit_point_tool():
    """Test the edit point tool for a CurveItem."""
    with qt_app_context(exec_loop=False):
        win, tool = create_window(EditPointTool)
        win.show()
        curve_item: CurveItem = win.manager.get_plot().get_active_item()  # type: ignore
        orig_x, orig_y = curve_item.get_data()

        assert orig_x is not None and orig_y is not None
        orig_x, orig_y = orig_x.copy(), orig_y.copy()

        assert tool is not None

        # must activate downsampling because the curve is too dense so the selection
        # distance threshold is too small for the programmed drag to select the first
        # point
        curve_item.param.dsamp_factor = 20
        win.manager.add_tool(DownSamplingTool).activate()
        # The steps must be very small to ensure the mouse passes close
        # enough to the first point of the curve to move it (distance < threshold)
        n = 1000
        min_v, max_v = 0, 1
        x_path = np.full(n, min_v)
        y_path = np.linspace(max_v, min_v, n)
        drag_mouse(win, x_path, y_path)

        x_path = np.full(n, max_v)

        drag_mouse(win, x_path, y_path)
        curve_changes = tool.get_changes()[curve_item]

        x_arr, y_arr = curve_item.get_data()
        for i, (x, y) in curve_changes.items():
            assert x == x_arr[i]
            assert y == y_arr[i]

        x_arr, y_arr = tool.get_arrays()
        assert x_arr is not None and y_arr is not None

        x_arr, y_arr = tool.get_initial_arrays()
        assert x_arr is not None and y_arr is not None

        # Reset the arrays and deletes the changes
        keyboard_event(win, QC.Qt.Key.Key_Z, QC.Qt.KeyboardModifier.ControlModifier)

        assert len(curve_changes) == 0

        restored_x, restored_y = curve_item.get_data()
        assert restored_x is not None and restored_y is not None
        assert np.allclose(orig_x, restored_x)
        assert np.allclose(orig_y, restored_y)

        mouse_event_at_relative_plot_pos(win, (0.5, 0.5), CLICK)
        tool.trigger_insert_point_at_selection()

        new_x, new_y = curve_item.get_data()
        assert len(new_x) == len(orig_x) + 1 and len(new_y) == len(orig_y) + 1

        tool.reset_arrays()
        assert tool.get_changes() == {}
        x_arr, y_arr = curve_item.get_data()
        assert np.isin(x_arr, orig_x).all() and np.isin(y_arr, orig_y).all()

        exec_dialog(win)


if __name__ == "__main__":
    test_curve_stat_tool()
    test_free_select_point_tool()
    test_contrained_select_point_tool()
    test_select_points_tool()
    test_edit_point_tool()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Test PlotBuilder image factory methods"""

import numpy as np
import pytest
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.constants import LUTAlpha
from plotpy.tests.unit.test_builder_curve import show_items_qtbot


def _make_image(
    alpha_function=None,
    xdata=[None, None],
    ydata=[None, None],
    pixel_size=None,
    center_on=None,
    interpolation="linear",
    background_color=None,
    eliminate_outliers=None,
    lut_range=None,
    lock_position=None,
):
    """Make image"""
    data = np.ones((50, 50), dtype=np.uint16)
    return make.image(
        data,
        title="Title",
        alpha_function=alpha_function,
        alpha=0.5,
        background_color=background_color,
        colormap="jet",
        xdata=xdata,
        ydata=ydata,
        pixel_size=pixel_size,
        center_on=center_on,
        interpolation=interpolation,
        eliminate_outliers=eliminate_outliers,
        xformat="%.1f",
        yformat="%.1f",
        zformat="%.1f",
        lut_range=lut_range,
        lock_position=lock_position,
    )


@pytest.mark.parametrize(
    "alpha_function",
    [
        None,
        LUTAlpha.NONE,
        LUTAlpha.CONSTANT,
        LUTAlpha.LINEAR,
        LUTAlpha.SIGMOID,
        LUTAlpha.TANH,
        LUTAlpha.STEP,
    ],
)
def test_builder_image_alpha_function(alpha_function):
    with qt_app_context(exec_loop=False):
        item = _make_image(alpha_function=alpha_function)
        show_items_qtbot([item])


@pytest.mark.parametrize(
    "xdata,ydata", [[[None, None], [None, None]], [[-10, 10], [-10, 10]]]
)
def test_builder_image_xdata_ydata(xdata, ydata):
    with qt_app_context(exec_loop=False):
        item = _make_image(xdata=xdata, ydata=ydata)
        show_items_qtbot([item])


@pytest.mark.parametrize("pixel_size", [None, 1.0, (1.0, 2.0)])
def test_builder_image_pixel_size(pixel_size):
    with qt_app_context(exec_loop=False):
        item = _make_image(pixel_size=pixel_size)
        show_items_qtbot([item])


@pytest.mark.parametrize("center_on", [None, [1.0, 3.0]])
def test_builder_image_center_on(center_on):
    with qt_app_context(exec_loop=False):
        item = _make_image(center_on=center_on, pixel_size=(1.0, 1.0))
        show_items_qtbot([item])


@pytest.mark.parametrize("interpolation", ["nearest", "linear", "antialiasing"])
def test_builder_image_interpolation(interpolation):
    with qt_app_context(exec_loop=False):
        item = _make_image(interpolation=interpolation)
        show_items_qtbot([item])


@pytest.mark.parametrize("background_color", [None, "red"])
def test_builder_image_background_color(background_color):
    with qt_app_context(exec_loop=False):
        item = _make_image(background_color=background_color)
        show_items_qtbot([item])


@pytest.mark.parametrize("eliminate_outliers", [None, 3.0])
def test_builder_image_eliminate_outliers(eliminate_outliers):
    with qt_app_context(exec_loop=False):
        item = _make_image(eliminate_outliers=eliminate_outliers)
        show_items_qtbot([item])


@pytest.mark.parametrize("lut_range", [None, [0.0, 100.0]])
def test_builder_image_lut_range(lut_range):
    with qt_app_context(exec_loop=False):
        item = _make_image(lut_range=lut_range)
        show_items_qtbot([item])


@pytest.mark.parametrize("lock_position", [None, True, False])
def test_builder_image_lock_position(lock_position):
    with qt_app_context(exec_loop=False):
        item = _make_image(lock_position=lock_position)
        show_items_qtbot([item])



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

from __future__ import annotations

import os
import os.path as osp
from typing import TYPE_CHECKING, Any

import numpy as np
from guidata.dataset import update_dataset
from guidata.utils.misc import assert_interfaces_valid
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qwt import QwtLinearScaleEngine, QwtPlotItem

from plotpy import io
from plotpy._scaler import (
    INTERP_AA,
    INTERP_LINEAR,
    INTERP_NEAREST,
    _histogram,
    _scale_rect,
)
from plotpy.config import _
from plotpy.constants import LUT_MAX, LUT_SIZE, LUTAlpha
from plotpy.coords import pixelround
from plotpy.interfaces import (
    IBaseImageItem,
    IBasePlotItem,
    IColormapImageItemType,
    ICSImageItemType,
    IExportROIImageItemType,
    IHistDataSource,
    IImageItemType,
    ISerializableType,
    ITrackableItemType,
    IVoiImageItemType,
)

# do not import rectangleshape from plotpy.items directly
from plotpy.items.shape.rectangle import RectangleShape
from plotpy.lutrange import lut_range_threshold
from plotpy.mathutils.arrayfuncs import get_nan_range
from plotpy.mathutils.colormap import FULLRANGE, get_cmap
from plotpy.styles.image import RawImageParam

if TYPE_CHECKING:
    import guidata.io
    import qwt.color_map
    import qwt.scale_map
    from qtpy.QtCore import QPointF, QRectF
    from qtpy.QtGui import QColor, QPainter

    from plotpy.interfaces import IItemType
    from plotpy.plot import BasePlot
    from plotpy.styles.base import ItemParameters
    from plotpy.widgets.colormap.widget import EditableColormap


class BaseImageItem(QwtPlotItem):
    """Base class for image items

    Args:
        data: Image data
        param: Image parameters
    """

    __implements__ = (
        IBasePlotItem,
        IBaseImageItem,
        IHistDataSource,
        IVoiImageItemType,
        ICSImageItemType,
        IExportROIImageItemType,
    )
    _can_select = True
    _can_resize = False
    _can_move = False
    _can_rotate = False
    _readonly = False
    _private = False
    _icon_name = "image.png"

    def __init__(
        self, data: np.ndarray | None = None, param: Any | None = None
    ) -> None:
        super().__init__()

        self.bg_qcolor = QG.QColor()

        self.bounds = QC.QRectF()

        # BaseImageItem needs:
        # param.background
        # param.alpha_function
        # param.alpha
        # param.colormap
        if param is None:
            param = self.get_default_param()
        self.param = param

        self.selected = False

        self.data: np.ndarray | None = None

        self.min = 0.0
        self.max = 1.0
        self.cmap_table = None
        self.cmap = None
        self.colormap_axis = None
        self._image: QG.QImage | None = None

        self._offscreen = np.array((1, 1), np.uint32)

        # Linear interpolation is the default interpolation algorithm:
        # it's almost as fast as 'nearest pixel' method but far smoother
        self.interpolate = None
        self.set_interpolation(INTERP_LINEAR)

        x1, y1 = self.bounds.left(), self.bounds.top()
        x2, y2 = self.bounds.right(), self.bounds.bottom()
        self.border_rect = RectangleShape(x1, y1, x2, y2)
        self.border_rect.set_style("plot", "shape/imageborder")
        # A, B, Background, Colormap
        self.lut = (1.0, 0.0, None, np.zeros((LUT_SIZE,), np.uint32))
        self.set_lut_range((0.0, 255.0))
        self.setItemAttribute(QwtPlotItem.AutoScale)
        self.setItemAttribute(QwtPlotItem.Legend, True)
        self._filename = None  # The file this image comes from

        self.histogram_cache = None
        if data is not None:
            self.set_data(data)
        self.param.update_item(self)

    # ---- Public API ----------------------------------------------------------
    def get_default_param(self) -> Any:
        """Return instance of the default image param DataSet"""
        raise NotImplementedError

    def set_filename(self, fname: str) -> None:
        """Set the filename of the image

        Args:
            fname: Filename
        """
        self._filename = fname

    def get_filename(self) -> str | None:
        """Get the filename of the image

        Returns:
            Filename
        """
        fname = self._filename
        if fname is not None and not osp.isfile(fname):
            other_try = osp.join(os.getcwd(), osp.basename(fname))
            if osp.isfile(other_try):
                self.set_filename(other_try)
                fname = other_try
        return fname

    def get_filter(self, filterobj, filterparam):
        """Provides a filter object over this image's content"""
        raise NotImplementedError

    def get_pixel_coordinates(self, xplot: float, yplot: float) -> tuple[float, float]:
        """Get pixel coordinates from plot coordinates

        Args:
            xplot: X plot coordinate
            yplot: Y plot coordinate

        Returns:
            Pixel coordinates
        """
        return xplot, yplot

    def get_plot_coordinates(self, xpixel: float, ypixel: float) -> tuple[float, float]:
        """Get plot coordinates from pixel coordinates

        Args:
            xpixel: X pixel coordinate
            ypixel: Y pixel coordinate

        Returns:
            Plot coordinates
        """
        return xpixel, ypixel

    def get_closest_indexes(
        self, x: float, y: float, corner: str | None = None
    ) -> tuple[int, int]:
        """Get closest image pixel indexes to the given coordinates

        Args:
            x: X coordinate
            y: Y coordinate
            corner: None (not a corner), 'TL' (top-left corner),
             'BR' (bottom-right corner)

        Returns:
            Closest image pixel indexes
        """
        x, y = self.get_pixel_coordinates(x, y)
        i_max = self.data.shape[1] - 1
        j_max = self.data.shape[0] - 1
        if corner == "BR":
            i_max += 1
            j_max += 1
        i = max([0, min([i_max, int(pixelround(x, corner))])])
        j = max([0, min([j_max, int(pixelround(y, corner))])])
        return i, j

    def get_closest_index_rect(
        self, x0: float, y0: float, x1: float, y1: float, avoid_empty: bool = True
    ) -> tuple[int, int, int, int]:
        """Get closest image rectangular pixel area index bounds, optionally
        avoid returning an empty rectangular area (return at least 1x1 pixel area)

        Args:
            x0: X coordinate of first point
            y0: Y coordinate of first point
            x1: X coordinate of second point
            y1: Y coordinate of second point
            avoid_empty: True to avoid returning an empty rectangular area.
             Defaults to True.

        Returns:
            Closest image rectangular pixel area index bounds compatible with
            direct array slicing (e.g., data[iy0:iy1, ix0:ix1])

        .. note::

            Avoid returning empty rectangular area (return 1x1 pixel area instead).
            Handle reversed/not-reversed Y-axis orientation.
            Indices are compatible with direct array slicing and match get_data
            behavior.
        """
        ix0, iy0 = self.get_closest_indexes(x0, y0, corner="TL")
        # Use TL to avoid double +1 addition
        ix1, iy1 = self.get_closest_indexes(x1, y1, corner="TL")
        if ix0 > ix1:
            ix1, ix0 = ix0, ix1
        if iy0 > iy1:
            iy1, iy0 = iy0, iy1
        # Add +1 to bottom-right indices for slice compatibility (matches get_data)
        ix1 = min(ix1 + 1, self.data.shape[1])
        iy1 = min(iy1 + 1, self.data.shape[0])
        if ix0 == ix1 and avoid_empty:
            ix1 = min(ix1 + 1, self.data.shape[1])
        if iy0 == iy1 and avoid_empty:
            iy1 = min(iy1 + 1, self.data.shape[0])
        return ix0, iy0, ix1, iy1

    def align_rectangular_shape(self, shape: RectangleShape) -> None:
        """Align rectangular shape to image pixels

        Args:
            shape: Shape to align
        """
        ix0, iy0, ix1, iy1 = self.get_closest_index_rect(*shape.get_rect())
        x0, y0 = self.get_plot_coordinates(ix0, iy0)
        x1, y1 = self.get_plot_coordinates(ix1, iy1)
        shape.set_rect(x0, y0, x1, y1)

    def get_closest_pixel_indexes(self, x: float, y: float) -> tuple[int, int]:
        """Get closest pixel indexes

        Args:
            x: X coordinate
            y: Y coordinate

        Returns:
            Closest pixel indexes

        .. note::

            Instead of returning indexes of an image pixel like the method
            'get_closest_indexes', this method returns the indexes of the
            closest pixel which is not necessarily on the image itself
            (i.e. indexes may be outside image index bounds: negative or
            superior than the image dimension)

        .. note::

            This is *not* the same as retrieving the canvas pixel coordinates
            (which depends on the zoom level)
        """
        x, y = self.get_pixel_coordinates(x, y)
        i = int(pixelround(x))
        j = int(pixelround(y))
        return i, j

    def get_x_values(self, i0: int, i1: int) -> np.ndarray:
        """Get X values from pixel indexes

        Args:
            i0: First index
            i1: Second index

        Returns:
            X values corresponding to the given pixel indexes
        """
        return np.arange(i0, i1)

    def get_y_values(self, j0: int, j1: int) -> np.ndarray:
        """Get Y values from pixel indexes

        Args:
            j0: First index
            j1: Second index

        Returns:
            Y values corresponding to the given pixel indexes
        """
        return np.arange(j0, j1)

    def get_r_values(self, i0, i1, j0, j1, flag_circle=False):
        """Get radial values from pixel indexes

        Args:
            i0: First index
            i1: Second index
            j0: Third index
            j1: Fourth index
            flag_circle: Flag circle (Default value = False)

        Returns:
            Radial values corresponding to the given pixel indexes
        """
        return self.get_x_values(i0, i1)

    def set_data(
        self, data: np.ndarray, lut_range: tuple[float, float] | None = None
    ) -> None:
        """Set image data

        Args:
            data: 2D NumPy array
            lut_range: LUT range -- tuple (levelmin, levelmax) (Default value = None)
        """
        self.data = data
        self.histogram_cache = None
        self.update_bounds()
        self.update_border()
        if not self.param.keep_lut_range:
            if lut_range is not None:
                _min, _max = lut_range
            else:
                _min, _max = get_nan_range(data)
            self.set_lut_range((_min, _max))

    def get_data(
        self, x0: float, y0: float, x1: float | None = None, y1: float | None = None
    ) -> float | tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Get image data

        Args:
            x0: X coordinate of first point
            y0: Y coordinate of first point
            x1: X coordinate of second point (Default value = None)
            y1: Y coordinate of second point (Default value = None)

        Returns:
            Image level, or image levels in rectangular area (x0,y0,x1,y1)
             as a tuple (x, y, data) of arrays
        """
        i0, j0 = self.get_closest_indexes(x0, y0)
        if x1 is None or y1 is None:
            return self.data[j0, i0]
        else:
            i1, j1 = self.get_closest_indexes(x1, y1)
            i1 += 1
            j1 += 1
            return (
                self.get_x_values(i0, i1),
                self.get_y_values(j0, j1),
                self.data[j0:j1, i0:i1],
            )

    def get_closest_coordinates(self, x: float, y: float) -> tuple[float, float]:
        """
        Get the closest coordinates to the given point

        Args:
            x: X coordinate
            y: Y coordinate

        Returns:
            tuple[float, float]: Closest coordinates
        """
        return self.get_closest_indexes(x, y)

    def get_coordinates_label(self, x: float, y: float) -> str:
        """
        Get the coordinates label for the given coordinates

        Args:
            x: X coordinate
            y: Y coordinate

        Returns:
            str: Coordinates label
        """
        title = self.title().text()
        z = self.get_data(x, y)
        zstr = "--" if z is np.ma.masked else f"{z:g}"
        return f"{title}:<br>x = {int(x):d}<br>y = {int(y):d}<br>z = {zstr}"

    def set_background_color(self, qcolor: QColor | str) -> None:
        """Set background color

        Args:
            qcolor: Background color
        """
        # mask = np.uint32(255*self.param.alpha+0.5).clip(0,255) << 24
        self.bg_qcolor = qcolor
        a, b, _bg, cmap = self.lut
        if qcolor is None:
            self.lut = (a, b, None, cmap)
        else:
            self.lut = (a, b, np.uint32(QG.QColor(qcolor).rgb() & 0xFFFFFF), cmap)

    def set_color_map(
        self, name_or_table: str | EditableColormap, invert: bool | None = None
    ) -> None:
        """Set colormap

        Args:
            name_or_table: Colormap name or colormap
            invert: True to invert colormap, False otherwise (Default value = None,
             i.e. do not change the default behavior)
        """
        if name_or_table is self.cmap_table:
            # This avoids rebuilding the LUT all the time
            return
        if isinstance(name_or_table, str):
            table = get_cmap(name_or_table)
        else:
            table = name_or_table
        if invert is not None:
            table.invert = invert
        self.cmap_table = table
        self.cmap = table.colorTable(FULLRANGE)
        cmap_a = self.lut[3]
        alpha = self.param.alpha
        alpha_function = self.param.alpha_function
        for i in range(LUT_SIZE):
            if alpha_function == LUTAlpha.NONE.value:
                pix_alpha = 1.0
            elif alpha_function == LUTAlpha.CONSTANT.value:
                pix_alpha = alpha
            else:
                x = i / float(LUT_SIZE - 1)
                if alpha_function == LUTAlpha.LINEAR.value:
                    # Linear alpha function
                    pix_alpha = alpha * x
                elif alpha_function == LUTAlpha.SIGMOID.value:
                    # Sigmoid alpha function
                    pix_alpha = alpha / (1 + np.exp(-10 * x))
                elif alpha_function == LUTAlpha.TANH.value:
                    # Hyperbolic tangent alpha function
                    pix_alpha = alpha * np.tanh(5 * x)
                elif alpha_function == LUTAlpha.STEP.value:
                    # Fully transparent lowest value and `alpha` transparent elsewhere
                    pix_alpha = alpha if x > 0 else 0
                else:
                    raise ValueError(f"Invalid alpha function {alpha_function}")
            # pylint: disable=too-many-function-args
            alpha_channel = max(min(np.uint32(255 * pix_alpha + 0.5), 255), 0) << 24
            cmap_a[i] = (
                np.uint32((table.rgb(FULLRANGE, i / LUT_MAX)) & 0xFFFFFF)
                | alpha_channel
            )
        plot = self.plot()
        if plot:
            plot.update_colormap_axis(self)

    def get_color_map(self) -> EditableColormap | None:
        """Get colormap

        Returns:
            Colormap
        """
        return self.cmap_table

    def set_interpolation(self, interp_mode: int, size: int | None = None) -> None:
        """Set interpolation mode

        Args:
            interp_mode: INTERP_NEAREST, INTERP_LINEAR, INTERP_AA
            size: (for anti-aliasing only) AA matrix size (Default value = None)
        """
        if interp_mode in (INTERP_NEAREST, INTERP_LINEAR):
            self.interpolate = (interp_mode,)
        if interp_mode == INTERP_AA:
            aa = np.ones((size, size), self.data.dtype)
            self.interpolate = (interp_mode, aa)

    def get_interpolation(self) -> tuple[int] | tuple[int, np.ndarray]:
        """Get interpolation mode"""
        return self.interpolate

    def set_lut_range(self, lut_range: tuple[float, float]) -> None:
        """
        Set the current active lut range

        Args:
            lut_range: Lut range, tuple(min, max)

        Example:
            >>> item.set_lut_range((0.0, 1.0))
        """
        self.min, self.max = lut_range
        _a, _b, bg, cmap = self.lut
        if self.max == self.min:
            self.lut = (LUT_MAX, self.min, bg, cmap)
        else:
            fmin, fmax = float(self.min), float(self.max)  # avoid overflows
            self.lut = (
                LUT_MAX / (fmax - fmin),
                -LUT_MAX * fmin / (fmax - fmin),
                bg,
                cmap,
            )

    def get_lut_range(self) -> tuple[float, float]:
        """Get the current active lut range

        Returns:
            tuple[float, float]: Lut range, tuple(min, max)
        """
        return self.min, self.max

    def set_lut_threshold(self, threshold: float) -> None:
        """Set lut threshold value, eliminating the given percent of the histogram

        Args:
            threshold: Lut threshold value
        """
        self.set_lut_range(lut_range_threshold(self, 256, threshold))

    def get_lut_range_full(self) -> tuple[float, float]:
        """Return full dynamic range

        Returns:
            tuple[float, float]: Lut range, tuple(min, max)
        """
        return get_nan_range(self.data)

    def get_lut_range_max(self) -> tuple[float, float]:
        """Get maximum range for this dataset

        Returns:
            tuple[float, float]: Lut range, tuple(min, max)
        """
        kind = self.data.dtype.kind
        if kind in np.typecodes["AllFloat"]:
            info = np.finfo(self.data.dtype)
        else:
            info = np.iinfo(self.data.dtype)
        return info.min, info.max

    def update_bounds(self) -> None:
        """Update image bounds to fit image shape"""
        if self.data is None:
            return
        self.bounds = QC.QRectF(0, 0, self.data.shape[1], self.data.shape[0])

    def update_border(self) -> None:
        """Update image border rectangle to fit image shape"""
        bounds = self.boundingRect().getCoords()
        self.border_rect.set_rect(*bounds)

    def draw_border(
        self,
        painter: QPainter,
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
        canvasRect: QRectF,
    ) -> None:
        """Draw image border rectangle

        Args:
            painter: Painter
            xMap: X axis scale map
            yMap: Y axis scale map
            canvasRect: Canvas rectangle
        """
        self.border_rect.draw(painter, xMap, yMap, canvasRect)

    def warn_if_non_linear_scale(self, painter: QPainter, canvasRect: QRectF) -> bool:
        """Warn if non-linear scale

        Args:
            painter: Painter
            canvasRect: Canvas rectangle

        Returns:
            True if non-linear scale
        """
        plot: BasePlot = self.plot()
        if not isinstance(
            plot.axisScaleEngine(plot.X_BOTTOM), QwtLinearScaleEngine
        ) or not isinstance(plot.axisScaleEngine(plot.Y_LEFT), QwtLinearScaleEngine):
            painter.setPen(QC.Qt.red)
            painter.drawText(
                canvasRect,
                QC.Qt.AlignCenter,
                _("Non-linear scales are not supported for image items"),
            )
            return True
        return False

    def draw_image(
        self,
        painter: QPainter,
        canvasRect: QRectF,
        src_rect: tuple[float, float, float, float],
        dst_rect: tuple[float, float, float, float],
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
    ) -> None:
        """Draw image

        Args:
            painter: Painter
            canvasRect: Canvas rectangle
            src_rect: Source rectangle
            dst_rect: Destination rectangle
            xMap: X axis scale map
            yMap: Y axis scale map
        """
        if self.warn_if_non_linear_scale(painter, canvasRect):
            return
        dest = _scale_rect(
            self.data, src_rect, self._offscreen, dst_rect, self.lut, self.interpolate
        )
        qrect = QC.QRectF(QC.QPointF(dest[0], dest[1]), QC.QPointF(dest[2], dest[3]))
        painter.drawImage(qrect, self._image, qrect)

    def export_roi(
        self,
        src_rect: tuple[float, float, float, float],
        dst_rect: tuple[float, float, float, float],
        dst_image: np.ndarray,
        apply_lut: bool = False,
        apply_interpolation: bool = False,
        original_resolution: bool = False,
        force_interp_mode: str | None = None,
        force_interp_size: int | None = None,
    ) -> None:
        """
        Export a rectangular area of the image to another image

        Args:
            src_rect: Source rectangle
            dst_rect: Destination rectangle
            dst_image: Destination image
            apply_lut: Apply lut (Default value = False)
            apply_interpolation: Apply interpolation (Default value = False)
            original_resolution: Original resolution (Default value = False)
            force_interp_mode: Force interpolation mode (Default value = None)
            force_interp_size: Force interpolation size (Default value = None)
        """
        if apply_lut:
            a, b, _bg, _cmap = self.lut
        else:
            a, b = 1.0, 0.0
        interp = self.interpolate if apply_interpolation else (INTERP_NEAREST,)
        _scale_rect(self.data, src_rect, dst_image, dst_rect, (a, b, None), interp)

    # ---- QwtPlotItem API -----------------------------------------------------
    def draw(
        self,
        painter: QPainter,
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
        canvasRect: QRectF,
    ) -> None:
        """Draw the item

        Args:
            painter: Painter
            xMap: X axis scale map
            yMap: Y axis scale map
            canvasRect: Canvas rectangle
        """
        x1, y1, x2, y2 = canvasRect.getCoords()
        i1, i2 = xMap.invTransform(x1), xMap.invTransform(x2)
        j1, j2 = yMap.invTransform(y1), yMap.invTransform(y2)

        xl, yt, xr, yb = self.boundingRect().getCoords()
        dest = (
            xMap.transform(xl),
            yMap.transform(yt),
            xMap.transform(xr) + 1,
            yMap.transform(yb) + 1,
        )

        W = int(canvasRect.right())
        H = int(canvasRect.bottom())
        if self._offscreen.shape != (H, W):
            self._offscreen = np.empty((H, W), np.uint32)
            self._image = QG.QImage(self._offscreen, W, H, QG.QImage.Format_ARGB32)
            self._image.ndarray = self._offscreen
            self.notify_new_offscreen()
        self.draw_image(painter, canvasRect, (i1, j1, i2, j2), dest, xMap, yMap)
        self.draw_border(painter, xMap, yMap, canvasRect)

    def boundingRect(self) -> QRectF:
        """Return the bounding rectangle of the data

        Returns:
            Bounding rectangle of the data
        """
        return self.bounds

    def notify_new_offscreen(self) -> None:
        """Notify that the offscreen image has changed"""
        # callback for those derived classes who need it
        pass

    def setVisible(self, enable: bool) -> None:
        """Set item visibility

        Args:
            enable: True if item is visible, False otherwise
        """
        if not enable:
            self.unselect()  # when hiding item, unselect it
        if enable:
            self.border_rect.show()
        else:
            self.border_rect.hide()
        QwtPlotItem.setVisible(self, enable)

    # ---- IBasePlotItem API ----------------------------------------------------
    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (
            IImageItemType,
            IVoiImageItemType,
            IColormapImageItemType,
            ITrackableItemType,
            ICSImageItemType,
            IExportROIImageItemType,
        )

    def set_readonly(self, state: bool) -> None:
        """Set object readonly state

        Args:
            state: True if object is readonly, False otherwise
        """
        self._readonly = state

    def is_readonly(self) -> bool:
        """Return object readonly state

        Returns:
            bool: True if object is readonly, False otherwise
        """
        return self._readonly

    def set_private(self, state: bool) -> None:
        """Set object as private

        Args:
            state: True if object is private, False otherwise
        """
        self._private = state

    def is_private(self) -> bool:
        """Return True if object is private

        Returns:
            bool: True if object is private, False otherwise
        """
        return self._private

    def select(self) -> None:
        """
        Select the object and eventually change its appearance to highlight the
        fact that it's selected
        """
        self.selected = True
        self.border_rect.select()

    def unselect(self) -> None:
        """
        Unselect the object and eventually restore its original appearance to
        highlight the fact that it's not selected anymore
        """
        self.selected = False
        self.border_rect.unselect()

    def is_empty(self) -> bool:
        """Return True if the item is empty

        Returns:
            True if the item is empty, False otherwise
        """
        return self.data is None or self.data.size == 0

    def get_icon_name(self) -> str:
        """Return the icon name

        Returns:
            Icon name
        """
        return self._icon_name

    def set_icon_name(self, icon_name: str) -> None:
        """Set the icon name

        Args:
            icon_name: Icon name
        """
        self._icon_name = icon_name

    def set_selectable(self, state: bool) -> None:
        """Set item selectable state

        Args:
            state: True if item is selectable, False otherwise
        """
        self._can_select = state

    def set_resizable(self, state: bool) -> None:
        """Set item resizable state
        (or any action triggered when moving an handle, e.g. rotation)

        Args:
            state: True if item is resizable, False otherwise
        """
        self._can_resize = state

    def set_movable(self, state: bool) -> None:
        """Set item movable state

        Args:
            state: True if item is movable, False otherwise
        """
        self._can_move = state

    def set_rotatable(self, state: bool) -> None:
        """Set item rotatable state

        Args:
            state: True if item is rotatable, False otherwise
        """
        self._can_rotate = state

    def can_select(self) -> bool:
        """
        Returns True if this item can be selected

        Returns:
            bool: True if item can be selected, False otherwise
        """
        return self._can_select

    def can_resize(self) -> bool:
        """
        Returns True if this item can be resized

        Returns:
            bool: True if item can be resized, False otherwise
        """
        return self._can_resize and not getattr(self.param, "lock_position", False)

    def can_move(self) -> bool:
        """
        Returns True if this item can be moved

        Returns:
            bool: True if item can be moved, False otherwise
        """
        return self._can_move and not getattr(self.param, "lock_position", False)

    def can_rotate(self) -> bool:
        """
        Returns True if this item can be rotated

        Returns:
            bool: True if item can be rotated, False otherwise
        """
        return self._can_rotate and not getattr(self.param, "lock_position", False)

    def hit_test(self, pos: QPointF) -> tuple[float, float, bool, None]:
        """Return a tuple (distance, attach point, inside, other_object)

        Args:
            pos: Position

        Returns:
            tuple: Tuple with four elements: (distance, attach point, inside,
             other_object).

        Description of the returned values:

        * distance: distance in pixels (canvas coordinates) to the closest
           attach point
        * attach point: handle of the attach point
        * inside: True if the mouse button has been clicked inside the object
        * other_object: if not None, reference of the object which will be
           considered as hit instead of self
        """
        plot = self.plot()
        ax = self.xAxis()
        ay = self.yAxis()
        return self.border_rect.poly_hit_test(plot, ax, ay, pos)

    def update_item_parameters(self) -> None:
        """Update item parameters (dataset) from object properties"""
        pass

    def get_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Appends datasets to the list of DataSets describing the parameters
        used to customize apearance of this item

        Args:
            itemparams: Item parameters
        """
        itemparams.add("ShapeParam", self, self.border_rect.shapeparam)

    def set_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Change the appearance of this item according
        to the parameter set provided

        Args:
            itemparams: Item parameters
        """
        self.border_rect.set_item_parameters(itemparams)

    def move_local_point_to(self, handle: int, pos: QPointF, ctrl: bool = None) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        pass

    def move_local_shape(self, old_pos: QPointF, new_pos: QPointF) -> None:
        """Translate the shape such that old_pos becomes new_pos in canvas coordinates

        Args:
            old_pos: Old position
            new_pos: New position
        """
        pass

    def move_with_selection(self, delta_x: float, delta_y: float) -> None:
        """Translate the item together with other selected items

        Args:
            delta_x: Translation in plot coordinates along x-axis
            delta_y: Translation in plot coordinates along y-axis
        """
        pass

    def resize_with_selection(self, zoom_dx, zoom_dy):
        """
        Resize the shape together with other selected items
        zoom_dx, zoom_dy : zoom factor for dx and dy
        """
        pass

    def rotate_with_selection(self, angle):
        """
        Rotate the shape together with other selected items
        angle : rotation angle
        """
        pass

    # ---- IBaseImageItem API --------------------------------------------------
    def can_sethistogram(self) -> bool:
        """
        Returns True if this item can be associated with a levels histogram

        Returns:
            bool: True if item can be associated with a levels histogram,
             False otherwise
        """
        return False

    def get_histogram(
        self, nbins: int, drange: tuple[float, float] | None = None
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Return a tuple (hist, bins) where hist is a list of histogram values

        Args:
            nbins: number of bins
            drange: lower and upper range of the bins. If not provided, range is
             simply (data.min(), data.max()). Values outside the range are ignored.

        Returns:
            Tuple (hist, bins)
        """
        if self.data is None:
            return [0], [0, 1]
        if self.histogram_cache is None or nbins != self.histogram_cache[0].shape[0]:
            if True:
                # Note: np.histogram does not accept data with NaN
                res = np.histogram(
                    self.data[~np.isnan(self.data)], bins=nbins, range=drange
                )
            else:
                # TODO: _histogram is faster, but caching is buggy in this version
                _min, _max = get_nan_range(self.data)
                if self.data.dtype in (np.float64, np.float32):
                    bins = np.unique(
                        np.array(
                            np.linspace(_min, _max, nbins + 1), dtype=self.data.dtype
                        )
                    )
                else:
                    bins = np.arange(_min, _max + 2, dtype=self.data.dtype)
                res2 = np.zeros((bins.size + 1,), np.uint32)
                _histogram(self.data.flatten(), bins, res2)
                res = res2[1:-1], bins
            self.histogram_cache = res
        else:
            res = self.histogram_cache
        return res

    def __process_cross_section(self, ydata, apply_lut):
        if apply_lut:
            a, b, _bg, _cmap = self.lut
            return (ydata * a + b).clip(0, LUT_MAX)
        else:
            return ydata

    def get_xsection(self, y0: float | int, apply_lut: bool = False) -> np.ndarray:
        """Return cross section along x-axis at y=y0

        Args:
            y0: Y0
            apply_lut: Apply lut (Default value = False)

        Returns:
            Cross section along x-axis at y=y0
        """
        _ix, iy = self.get_closest_indexes(0, y0)
        return (
            self.get_x_values(0, self.data.shape[1]),
            self.__process_cross_section(self.data[iy, :], apply_lut),
        )

    def get_ysection(self, x0: float | int, apply_lut: bool = False) -> np.ndarray:
        """Return cross section along y-axis at x=x0

        Args:
            x0: X0
            apply_lut: Apply lut (Default value = False)

        Returns:
            Cross section along y-axis at x=x0
        """
        ix, _iy = self.get_closest_indexes(x0, 0)
        return (
            self.get_y_values(0, self.data.shape[0]),
            self.__process_cross_section(self.data[:, ix], apply_lut),
        )

    def get_average_xsection(
        self, x0: float, y0: float, x1: float, y1: float, apply_lut: bool = False
    ) -> np.ndarray:
        """Return average cross section along x-axis for the given rectangle

        Args:
            x0: X0 of top left corner
            y0: Y0 of top left corner
            x1: X1 of bottom right corner
            y1: Y1 of bottom right corner
            apply_lut: Apply lut (Default value = False)

        Returns:
            Average cross section along x-axis
        """
        ix0, iy0, ix1, iy1 = self.get_closest_index_rect(
            x0, y0, x1, y1, avoid_empty=False
        )
        ydata = self.data[iy0:iy1, ix0:ix1]
        if ydata.size == 0:
            return np.array([]), np.array([])
        ydata = ydata.mean(axis=0)
        return (
            self.get_x_values(ix0, ix1),
            self.__process_cross_section(ydata, apply_lut),
        )

    def get_average_ysection(
        self, x0: float, y0: float, x1: float, y1: float, apply_lut: bool = False
    ) -> np.ndarray:
        """Return average cross section along y-axis

        Args:
            x0: X0 of top left corner
            y0: Y0 of top left corner
            x1: X1 of bottom right corner
            y1: Y1 of bottom right corner
            apply_lut: Apply lut (Default value = False)

        Returns:
            Average cross section along y-axis
        """
        ix0, iy0, ix1, iy1 = self.get_closest_index_rect(
            x0, y0, x1, y1, avoid_empty=False
        )
        ydata = self.data[iy0:iy1, ix0:ix1]
        if ydata.size == 0:
            return np.array([]), np.array([])
        ydata = ydata.mean(axis=1)
        return (
            self.get_y_values(iy0, iy1),
            self.__process_cross_section(ydata, apply_lut),
        )


assert_interfaces_valid(BaseImageItem)


# ==============================================================================
# Raw Image item (image item without scale)
# ==============================================================================
class RawImageItem(BaseImageItem):
    """
    Construct a simple image item

        * data: 2D NumPy array
        * param: image parameters
          (:py:class:`.styles.RawImageParam` instance)
    """

    __implements__ = (
        IBasePlotItem,
        IBaseImageItem,
        IHistDataSource,
        IVoiImageItemType,
        ISerializableType,
    )

    # ---- BaseImageItem API ---------------------------------------------------
    def get_default_param(self) -> RawImageParam:
        """Return instance of the default image param DataSet

        Returns:
            RawImageParam: Default image param DataSet
        """
        return RawImageParam(_("Image"))

    # ---- Serialization methods -----------------------------------------------
    def __reduce__(self) -> tuple:
        """Return state information for pickling"""
        fname = self.get_filename()
        if fname is None:
            fn_or_data = self.data
        else:
            fn_or_data = fname
        state = self.param, self.get_lut_range(), fn_or_data, self.z()
        res = (self.__class__, (), state)
        return res

    def __setstate__(self, state: tuple) -> None:
        """Restore state information for pickling"""
        param, lut_range, fn_or_data, z = state
        self.param = param
        if isinstance(fn_or_data, str):
            self.set_filename(fn_or_data)
            self.load_data()
        elif fn_or_data is not None:  # should happen only with previous API
            self.set_data(fn_or_data)
        self.set_lut_range(lut_range)
        self.setZ(z)
        self.param.update_item(self)

    def serialize(
        self,
        writer: guidata.io.HDF5Writer | guidata.io.INIWriter | guidata.io.JSONWriter,
    ) -> None:
        """Serialize object to HDF5 writer

        Args:
            writer: HDF5, INI or JSON writer
        """
        fname = self.get_filename()
        load_from_fname = fname is not None
        data = None if load_from_fname else self.data
        writer.write(load_from_fname, group_name="load_from_fname")
        writer.write(fname, group_name="fname")
        writer.write(data, group_name="Zdata")
        writer.write(self.get_lut_range(), group_name="lut_range")
        writer.write(self.z(), group_name="z")
        self.param.update_param(self)
        writer.write(self.param, group_name="imageparam")

    def deserialize(
        self,
        reader: guidata.io.HDF5Reader | guidata.io.INIReader | guidata.io.JSONReader,
    ) -> None:
        """Deserialize object from HDF5 reader

        Args:
            reader: HDF5, INI or JSON reader
        """
        lut_range = reader.read(group_name="lut_range")
        if reader.read(group_name="load_from_fname"):
            name = reader.read(group_name="fname", func=reader.read_unicode)
            self.set_filename(name)
            self.load_data()
        else:
            data = reader.read(group_name="Zdata", func=reader.read_array)
            self.set_data(data)
        self.set_lut_range(lut_range)
        self.setZ(reader.read("z"))
        self.param = self.get_default_param()
        reader.read("imageparam", instance=self.param)
        self.param.update_item(self)

    # ---- Public API ----------------------------------------------------------
    def load_data(self, lut_range: list[float, float] | None = None) -> None:
        """Load data from item filename attribute
        and eventually apply specified lut_range

        Args:
            lut_range: Lut range (Default value = None)
        """
        data = io.imread(self.get_filename(), to_grayscale=True)
        self.set_data(data, lut_range=lut_range)

    # ---- IBasePlotItem API ---------------------------------------------------
    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (
            IImageItemType,
            IVoiImageItemType,
            IColormapImageItemType,
            ITrackableItemType,
            ICSImageItemType,
            ISerializableType,
            IExportROIImageItemType,
        )

    def update_item_parameters(self) -> None:
        """Update item parameters (dataset) from object properties"""
        self.param.update_param(self)

    def get_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Appends datasets to the list of DataSets describing the parameters
        used to customize apearance of this item

        Args:
            itemparams: Item parameters
        """
        BaseImageItem.get_item_parameters(self, itemparams)
        self.update_item_parameters()
        itemparams.add("ImageParam", self, self.param)

    def set_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Change the appearance of this item according
        to the parameter set provided

        Args:
            itemparams: Item parameters
        """
        update_dataset(self.param, itemparams.get("ImageParam"), visible_only=True)
        self.param.update_item(self)
        BaseImageItem.set_item_parameters(self, itemparams)

    # ---- IBaseImageItem API --------------------------------------------------
    def can_sethistogram(self) -> bool:
        """
        Returns True if this item can be associated with a levels histogram

        Returns:
            bool: True if item can be associated with a levels histogram,
             False otherwise
        """
        return True


assert_interfaces_valid(RawImageItem)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Test showing 10 big images"""

# guitest: show

import numpy as np
import pytest
from guidata.qthelpers import qt_app_context

from plotpy.builder import make


def imshow():
    win = make.dialog(
        toolbar=True, title="Displaying 10 big images test", size=(800, 600)
    )
    plot = win.manager.get_plot()
    for i in range(10):
        plot.add_item(make.image(compute_image(i)))
    win.show()
    return win


def compute_image(i, N=7500, M=1750):
    if i % 2 == 0:
        N, M = M, N
    return (np.random.rand(N, M) * 65536).astype(np.int16)


@pytest.mark.skip(reason="Not relevant in automated test suite")
def test_bigimages():
    """Test Bigimages"""
    with qt_app_context(exec_loop=True):
        _persist_obj = imshow()


if __name__ == "__main__":
    test_bigimages()



# -*- coding: utf-8 -*-

"""
Colormaps
=========

The `colormap` module contains definition of common colormaps and tools
to manipulate and create them.

This package provides widgets for creating, editing, and managing colormaps:

- ``ColorMapWidget``: Interactive colormap editor widget
- ``ColorMapEditor``: Full colormap editing dialog
- ``ColorMapManager``: Colormap management interface
"""



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Unit tests for geometry module
"""

from __future__ import annotations

import numpy as np

from plotpy.mathutils import geometry as geom


def test_transform_matrix() -> None:
    """Testing transform matrix functions"""
    x0 = 1.0
    y0 = 2.0
    scale_x = 2.0
    scale_y = 3.0
    delta_x = 15.3
    delta_y = 10.4
    angle = 30.0
    vect0 = geom.colvector(x0, y0)
    # print("vect0:", repr(vect0))
    vect1 = geom.scale(scale_x, scale_y) @ vect0
    # print("vect1:", repr(vect1))
    assert np.allclose(vect1, np.array([2.0, 6.0, 1.0]))
    vect2 = geom.translate(delta_x, delta_y) @ vect1
    # print("vect2:", repr(vect2))
    assert np.allclose(vect2, np.array([17.3, 16.4, 1.0]))
    vect3 = geom.rotate(angle) @ vect2
    # print("vect3:", repr(vect3))
    assert np.allclose(vect3, np.array([18.87226872, -14.56322332, 1.0]))
    trmat = (
        geom.scale(1.0 / scale_x, 1.0 / scale_y)
        @ geom.translate(-delta_x, -delta_y)
        @ geom.rotate(-angle)
    )
    vect4 = trmat @ vect3
    # print("vect4:", repr(vect4))
    assert np.allclose(vect0, vect4)


def test_vector_operations() -> None:
    """Test vector operation functions"""
    xa = 1.0
    ya = 2.0
    xb = 3.0
    yb = 4.0
    vect = np.array([5.0, 6.0])
    angle = np.deg2rad(30.0)
    norm = geom.vector_norm(xa, ya, xb, yb)
    assert np.allclose(norm, np.sqrt(8.0))
    proj = geom.vector_projection(vect, xa, ya, xb, yb)
    # print("proj:", repr(proj))
    assert np.allclose(proj, np.array([8.5, 9.5]))
    vrot = geom.vector_rotation(angle, vect[0], vect[1])
    # print("vrot:", repr(vrot))
    assert np.allclose(vrot, np.array([1.33012702, 7.69615242]))
    angle = geom.vector_angle(vect[0], vect[1])
    # print("angle:", repr(angle))
    assert np.allclose(angle, np.arctan2(vect[1], vect[0]))


def test_coordinates_computations() -> None:
    """Testing coordinates computation functions"""
    x1 = 1.0
    y1 = 2.0
    x2 = 3.0
    y2 = 4.0
    xc, yc = geom.compute_center(x1, y1, x2, y2)
    assert np.allclose(xc, 2.0) and np.allclose(yc, 3.0)
    width, height = geom.compute_rect_size(x1, y1, x2, y2)
    assert np.allclose(width, 2.0) and np.allclose(height, 2.0)
    dist = geom.compute_distance(x1, y1, x2, y2)
    assert np.allclose(dist, np.sqrt(8.0))
    angle = geom.compute_angle(x1, y1, x2, y2)
    assert np.allclose(angle, np.rad2deg(-np.arctan2(2.0, 2.0)))
    angle_inv = geom.compute_angle(x1, y1, x2, y2, True)
    assert np.allclose(angle_inv, np.rad2deg(np.arctan2(2.0, 2.0)))


if __name__ == "__main__":
    test_transform_matrix()
    test_vector_operations()
    test_coordinates_computations()



# -*- coding: utf-8 -*-

"""
Basic geometry functions
------------------------

Overview
^^^^^^^^

The :py:mod:`.geometry` module provides basic geometry functions for
computing 2D transformations and distances.

The following functions are available:

* :py:func:`.translate`
* :py:func:`.scale`
* :py:func:`.rotate`
* :py:func:`.colvector`
* :py:func:`.vector_norm`
* :py:func:`.vector_projection`
* :py:func:`.vector_angle`
* :py:func:`.compute_center`
* :py:func:`.compute_rect_size`
* :py:func:`.compute_distance`
* :py:func:`.compute_angle`

Reference
^^^^^^^^^

.. autofunction:: translate
.. autofunction:: scale
.. autofunction:: rotate
.. autofunction:: colvector
.. autofunction:: vector_norm
.. autofunction:: vector_projection
.. autofunction:: vector_angle
.. autofunction:: compute_center
.. autofunction:: compute_rect_size
.. autofunction:: compute_distance
.. autofunction:: compute_angle
"""

# pylint: disable=C0103

from __future__ import annotations

import numpy as np

# ===============================================================================
# Transform matrix functions
# ===============================================================================


def translate(tx: float, ty: float) -> np.ndarray:
    """Return translation matrix

    Args:
        tx: Translation along X-axis
        ty: Translation along Y-axis

    Returns:
        Translation matrix
    """
    return np.array([[1, 0, tx], [0, 1, ty], [0, 0, 1]], float)


def scale(sx: float, sy: float) -> np.ndarray:
    """Return scale matrix

    Args:
        sx: Scale along X-axis
        sy: Scale along Y-axis

    Returns:
        Scale matrix
    """
    return np.array([[sx, 0, 0], [0, sy, 0], [0, 0, 1]], float)


def rotate(alpha: float) -> np.ndarray:
    """Return rotation matrix

    Args:
        alpha: Rotation angle (in radians)

    Returns:
        Rotation matrix
    """
    return np.array(
        [
            [np.cos(alpha), -np.sin(alpha), 0],
            [np.sin(alpha), np.cos(alpha), 0],
            [0, 0, 1],
        ],
        float,
    )


def colvector(x: float, y: float) -> np.ndarray:
    """Return vector from coordinates

    Args:
        x: x-coordinate
        y: y-coordinate

    Returns:
        Vector
    """
    return np.array([x, y, 1]).T


# ===============================================================================
# Operations on vectors (from coordinates)
# ===============================================================================


def vector_norm(xa: float, ya: float, xb: float, yb: float) -> float:
    """Return vector norm

    Args:
        xa: x-coordinate of first point
        ya: y-coordinate of first point
        xb: x-coordinate of second point
        yb: y-coordinate of second point

    Returns:
        Norm of vector (xa, xb)-->(ya, yb)
    """
    return np.linalg.norm(np.array((xb - xa, yb - ya)))


def vector_projection(
    dv: np.ndarray, xa: float, ya: float, xb: float, yb: float
) -> np.ndarray:
    """Return vector projection

    Args:
        dv: vector to project
        xa: x-coordinate of first point
        ya: y-coordinate of first point
        xb: x-coordinate of second point
        yb: y-coordinate of second point

    Returns:
        Projection of *dv* on vector (xa, xb)-->(ya, yb)
    """
    assert dv.shape == (2,)
    v_ab = np.array((xb - xa, yb - ya))
    u_ab = v_ab / np.linalg.norm(v_ab)
    return np.vdot(u_ab, dv) * u_ab + np.array((xb, yb))


def vector_rotation(theta: float, dx: float, dy: float) -> tuple[float, float]:
    """Compute theta-rotation on vector

    Args:
        theta: Rotation angle
        dx: x-coordinate of vector
        dy: y-coordinate of vector

    Returns:
        Tuple of (x, y) coordinates of rotated vector
    """
    return (rotate(theta) @ colvector(dx, dy)).ravel()[:2]


def vector_angle(dx: float, dy: float) -> float:
    """Return vector angle with X-axis

    Args:
        dx: x-coordinate of vector
        dy: y-coordinate of vector

    Returns:
        Angle between vector and X-axis (in radians)
    """
    # sign(dy) ==  1 --> return Arccos()
    # sign(dy) ==  0 --> return  0 if sign(dx) ==  1
    # sign(dy) ==  0 --> return pi if sign(dx) == -1
    # sign(dy) == -1 --> return 2pi-Arccos()
    if dx == 0 and dy == 0:
        return 0.0
    else:
        sx, sy = np.sign(dx), np.sign(dy)
        acos = np.arccos(dx / np.sqrt(dx**2 + dy**2))
        return sy * (np.pi * (sy - 1) + acos) + np.pi * (1 - sy**2) * (1 - sx) * 0.5


# ===============================================================================
# Misc.
# ===============================================================================


def compute_center(x1: float, y1: float, x2: float, y2: float) -> tuple[float, float]:
    """Compute center of rectangle

    Args:
        x1: x-coordinate of top-left corner
        y1: y-coordinate of top-left corner
        x2: x-coordinate of bottom-right corner
        y2: y-coordinate of bottom-right corner

    Returns:
        Tuple of (x, y) coordinates of center
    """
    return 0.5 * (x1 + x2), 0.5 * (y1 + y2)


def compute_rect_size(
    x1: float, y1: float, x2: float, y2: float
) -> tuple[float, float]:
    """Compute rectangle size

    Args:
        x1: x-coordinate of top-left corner
        y1: y-coordinate of top-left corner
        x2: x-coordinate of bottom-right corner
        y2: y-coordinate of bottom-right corner

    Returns:
        Tuple of (width, height)
    """
    return x2 - x1, np.fabs(y2 - y1)


def compute_distance(x1: float, y1: float, x2: float, y2: float) -> float:
    """Compute distance between two points

    Args:
        x1: x-coordinate of first point
        y1: y-coordinate of first point
        x2: x-coordinate of second point
        y2: y-coordinate of second point

    Returns:
        Distance between points
    """
    return np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)


def compute_angle(
    x1: float, y1: float, x2: float, y2: float, reverse: bool = False
) -> float:
    """Compute angle between two points

    Args:
        x1: x-coordinate of first point
        y1: y-coordinate of first point
        x2: x-coordinate of second point
        y2: y-coordinate of second point
        reverse: If True, return the angle in the opposite direction

    Returns:
        Angle between points (in degrees)
    """
    sign = -1 if reverse else 1
    if x2 == x1:
        return 0.0
    else:
        return np.arctan(-sign * (y2 - y1) / (x2 - x1)) * 180.0 / np.pi



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Colormap editor
---------------

This module provides a more complete colormap editor widget than the one provided
by ColorMapWidget (plotpy/widgets/colormap_widget.py). It allows to edit a colormap
by changing its color stops (add/delete/move/change color).

Reference
~~~~~~~~~

.. autoclass:: ColorPickDataSet
    :members:
.. autoclass:: ColorMapEditor
    :members:
"""

from __future__ import annotations

import qtpy.QtCore as QC
import qtpy.QtGui as QG
import qtpy.QtWidgets as QW
from guidata.dataset import ColorItem, DataSet, FloatItem
from guidata.dataset.datatypes import GetAttrProp, NotProp
from guidata.dataset.qtwidgets import DataSetEditGroupBox

from plotpy.config import _
from plotpy.widgets.colormap.widget import ColorMapWidget, EditableColormap


class ColorPickDataSet(DataSet):
    """Dataset with the field used to edit a color stop."""

    _position_locked = False

    position = (
        FloatItem(
            _("Relative position"),
            min=0.0,
            max=1.0,
            default=0.5,
            help=_("Must be a value between 0.0 and 1.0."),
        )
        .set_prop("display", format="%.2f")
        .set_prop("display", active=NotProp(GetAttrProp("_position_locked")))
    )
    hex_color = ColorItem(
        "Pick color", help="Pick a color to use for the selected cursor."
    )

    def lock_position(self, lock=True) -> None:
        """Used to lock the position of the cursor value.

        Args:
            lock: True if value must be locked. Defaults to True.
        """
        self._position_locked = lock

    def is_position_locked(self):
        """Returns True if the position is locked.

        Returns:
            True if position is locked, False otherwise.
        """
        return self._position_locked

    def set_position(self, position: float) -> None:
        """Set a new position for the cursor. Value is rounded to 2 decimals.

        Args:
            position: new position to set
        """
        self.position = round(position, 2)

    def get_position(self) -> float:
        """Return the current position of the cursor.

        Returns:
            float value of the cursor position
        """
        return self.position  # type: ignore

    def set_color(self, hex_color: str):
        """Set a new hex color for the cursor.

        Args:
            hex_color: str hex color to set (e.g. "#FF0000" for red)
        """
        self.hex_color = hex_color

    def get_hex_color(self) -> str:
        """Return the current str hex color of the cursor.

        Returns:
            str hex color of the cursor (e.g. "#FF0000" for red)
        """
        return self.hex_color  # type: ignore


class ColorMapEditor(QW.QWidget):
    """Widget that allows a user to edit a colormap by changing its color stops (
    add/delete/move/change color). Right click on the colorbar or slider to add or
    remove colors stops. A existing colormap instance can be used. However, the
    modifications are inplace so you should copy the object if you do not want to
    mutate it.

    Args:
        parent: Parent widget.
        cmap_width: Cmap size in pixels. Defaults to 400.
        cmap_height: Cmap height in pixels. Defaults to 50.
        color1: First color of the colormap. If None, color blue is used. Ignored if
         argument 'colormap' is used. Defaults to None.
        color2: Last color of the colormap. If None, color yellow is used.Ignored if
         argument 'colormap' is used. Defaults to None
        colormap: An already initialized colormap to use in the widget.
         Defaults to None.
    """

    def __init__(
        self,
        parent: QW.QWidget | None,
        cmap_width: int = 400,
        cmap_height: int = 50,
        color1: QG.QColor | None = None,
        color2: QG.QColor | None = None,
        colormap: EditableColormap | None = None,
    ) -> None:
        super().__init__(parent)

        color1 = QG.QColor(QC.Qt.GlobalColor.blue) if color1 is None else color1
        color2 = QG.QColor(QC.Qt.GlobalColor.yellow) if color2 is None else color2
        self.colormap_widget = ColorMapWidget(
            self, cmap_width, cmap_height, color1, color2, colormap
        )

        self.tabs = QW.QTabWidget()

        self.datasets: list[DataSetEditGroupBox[ColorPickDataSet]] = []

        self.setup_tabs()
        self.tabs.setCurrentIndex(0)
        self.update_tabs_names_from_current()

        layout = QW.QVBoxLayout(self)
        layout.addWidget(self.colormap_widget)
        layout.addWidget(self.tabs)
        layout.addStretch(1)
        self.setLayout(layout)

        self.colormap_widget.HANDLE_SELECTED.connect(self.change_current_dataset)
        self.colormap_widget.HANDLE_ADDED.connect(self.new_tab)
        self.colormap_widget.HANDLE_DELETED.connect(self.delete_tab)
        self.colormap_widget.multi_range_hslider.sliderMoved.connect(
            lambda *args: self.update_current_dataset()
        )

    def set_colormap(self, colormap: EditableColormap) -> None:
        """Replaces the current colormap.

        Args:
            colormap: replacement colormap
        """
        self.colormap_widget.blockSignals(True)
        self.colormap_widget.set_colormap(colormap)
        self.colormap_widget.blockSignals(False)
        self.setup_tabs()

    def get_colormap(self) -> EditableColormap:
        """Get the current colormap being edited.

        Returns:
            current colormap
        """
        return self.colormap_widget.get_colormap()

    def change_current_dataset(self, dataset_index: int) -> None:
        """Wrapper functio to change the current dataset (=current tab)

        Args:
            dataset_index: dataset index (=tab) to set current
        """
        self.tabs.setCurrentIndex(dataset_index)

    def update_tab_color(self, tab_index: int) -> None:
        """Update the tab icon color for the given tab index.

        Args:
            tab_index: index of the tab to update
        """
        handle_color = self.colormap_widget.get_handle_color(tab_index)

        pixmap = QG.QPixmap(10, 10)  # Size of the square
        pixmap.fill(handle_color)  # Color of the square
        new_icon = QG.QIcon(pixmap)
        self.tabs.setTabIcon(tab_index, new_icon)

        hex_color = handle_color.name()
        dataset_grp = self.datasets[tab_index]
        dataset_grp.dataset.set_color(hex_color)
        dataset_grp.get()

    def update_tabs_names_from_current(self) -> None:
        """Update all the tab names from the current active tab (useful when a new one
        is added).
        """
        start_index = self.tabs.currentIndex()
        for i in range(start_index, self.tabs.count()):
            self.tabs.setTabText(i, str(i + 1))

    def new_tab(self, index: int, handle_pos: float) -> None:
        """Add/insert a new tab at the given index and set its relative position and
        color.

        Args:
            index: index of the insertion/appending.
            handle_pos: relative value to set in the tab (new handle current position)
        """
        dw = DataSetEditGroupBox("", ColorPickDataSet)
        dw.dataset.set_position(handle_pos)

        hex_color = self.colormap_widget.get_hex_color(index)
        dw.dataset.set_color(hex_color)

        dw.SIG_APPLY_BUTTON_CLICKED.connect(self.update_colormap_widget)

        if index == self.tabs.count():
            self.datasets.append(dw)
            self.tabs.addTab(dw, "")
        else:
            self.datasets.insert(index, dw)
            self.tabs.insertTab(index, dw, "")
            dw.updateGeometry()

        self.tabs.setCurrentIndex(index)
        dw.set()

        self.update_tab_color(index)
        self.update_tabs_names_from_current()

    def setup_tabs(self) -> None:
        """Clear and setup all the tabs from the current colormap. Can be used to reset
        all tabs after initialization.
        """
        self.datasets.clear()
        self.tabs.clear()
        for i, handle_pos in enumerate(self.colormap_widget.get_handles_tuple()):
            self.new_tab(i, handle_pos)
        self.datasets[0].dataset.lock_position()
        self.datasets[0].get()
        self.datasets[-1].dataset.lock_position()
        self.datasets[-1].get()

    def reset_colors(self):
        """Updates all the tab icons color."""
        for i in range(len(self.colormap_widget)):
            self.update_tab_color(i)

    def delete_tab(self, tab_index: int) -> None:
        """Removes the tab/dataset at given index.

        Args:
            tab_index: index to remove
        """
        self.tabs.removeTab(tab_index)
        self.datasets.pop(tab_index)
        self.tabs.setCurrentIndex(tab_index)
        self.update_tabs_names_from_current()

    def update_colormap_widget(self) -> None:
        """Update the colormap widget (colorbar and handles) with the values from the
        current tab.
        """
        current_index = self.tabs.currentIndex()
        current_dataset = self.datasets[current_index].dataset

        new_slider_values = self.colormap_widget.get_handles_list()

        if 0 < current_index < self.tabs.count() - 1:
            relative_pos = current_dataset.get_position()
        else:
            relative_pos = new_slider_values[current_index]
            current_dataset.set_position(relative_pos)
            self.datasets[current_index].get()

        new_color = QG.QColor(current_dataset.get_hex_color())
        self.colormap_widget.edit_color_stop(current_index, relative_pos, new_color)
        self.update_tab_color(current_index)

    def update_current_dataset(self):
        """Update values in the current dataset/tab with the values from the colormap
        widget.
        """
        current_index = self.tabs.currentIndex()
        current_dataset_grp = self.datasets[current_index]
        relative_pos = self.colormap_widget.get_handles_tuple()[current_index]
        current_dataset_grp.dataset.set_position(relative_pos)
        current_dataset_grp.get()
        current_dataset_grp.get()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

# pylint: disable=C0103

"""
Shape Item builder
------------------

This module provides a set of factory functions to simplify the creation
of shape items.
"""

# Note: when adding method to builder classes, please do not forget to update the
# documentation (see builder.rst file). Because of class inheritance, the methods
# are not automatically documented (otherwise, they would be sorted alphabetically,
# due to a limitation of sphinx auto-doc).

from __future__ import annotations

import numpy as np

from plotpy.items import (
    CircleSVGShape,
    EllipseShape,
    PolygonShape,
    RectangleShape,
    RectangleSVGShape,
    SegmentShape,
    SquareSVGShape,
)


class ShapeBuilder:
    """Class regrouping a set of factory functions to simplify the creation
    of shape items."""

    def __shape(self, shapeclass, x0, y0, x1, y1, title=None):
        shape = shapeclass(x0, y0, x1, y1)
        shape.set_style("plot", "shape/drag")
        if title is not None:
            shape.setTitle(title)
        return shape

    def rectangle(
        self, x0: float, y0: float, x1: float, y1: float, title: str | None = None
    ) -> RectangleShape:
        """Make a rectangle shape `plot item`

        Args:
            x0: rectangle x0 coordinate
            y0: rectangle y0 coordinate
            x1: rectangle x1 coordinate
            y1: rectangle y1 coordinate
            title: label name. Default is None

        Returns:
            :py:class:`.RectangleShape` object
        """
        return self.__shape(RectangleShape, x0, y0, x1, y1, title)

    def ellipse(
        self,
        x0: float,
        y0: float,
        x1: float,
        y1: float,
        x2: float | None = None,
        y2: float | None = None,
        x3: float | None = None,
        y3: float | None = None,
        title: str | None = None,
    ) -> EllipseShape:
        """Make an ellipse shape `plot item`

        Args:
            x0: ellipse x0 coordinate
            y0: ellipse y0 coordinate
            x1: ellipse x1 coordinate
            y1: ellipse y1 coordinate
            x2: ellipse x2 coordinate. Default is None
            y2: ellipse y2 coordinate. Default is None
            x3: ellipse x3 coordinate. Default is None
            y3: ellipse y3 coordinate. Default is None
            title: label name. Default is None

        Returns:
            :py:class:`.EllipseShape` object
        """
        item = self.__shape(EllipseShape, x0, y0, x1, y1, title)
        item.switch_to_ellipse()
        if x2 is not None and y2 is not None and x3 is not None and y3 is not None:
            item.set_ydiameter(x2, y2, x3, y3)
        return item

    def polygon(
        self, x: np.ndarray, y: np.ndarray, closed: bool, title: str | None = None
    ) -> PolygonShape:
        """Make a polygon shape `plot item`

        Args:
            x: polygon x coordinates
            y: polygon y coordinates
            closed: closed polygon
            title: label name. Default is None

        Returns:
            :py:class:`.PolygonShape` object
        """
        points = np.array([x, y]).T
        item = PolygonShape(points, closed=closed)
        item.set_style("plot", "shape/drag")
        if title is not None:
            item.setTitle(title)
        return item

    def circle(
        self, x0: float, y0: float, x1: float, y1: float, title: str | None = None
    ) -> EllipseShape:
        """Make a circle shape `plot item`

        Args:
            x0: circle x0 coordinate
            y0: circle y0 coordinate
            x1: circle x1 coordinate
            y1: circle y1 coordinate
            title: label name. Default is None

        Returns:
            :py:class:`.EllipseShape` object
        """
        item = self.__shape(EllipseShape, x0, y0, x1, y1, title)
        item.switch_to_circle()
        return item

    def segment(
        self, x0: float, y0: float, x1: float, y1: float, title: str | None = None
    ) -> SegmentShape:
        """Make a segment shape `plot item`

        Args:
            x0: segment x0 coordinate
            y0: segment y0 coordinate
            x1: segment x1 coordinate
            y1: segment y1 coordinate
            title: label name. Default is None

        Returns:
            :py:class:`.SegmentShape` object
        """
        return self.__shape(SegmentShape, x0, y0, x1, y1, title)

    def svg(
        self,
        shape: str,
        fname_or_data: str | bytes,
        x0: float,
        y0: float,
        x1: float,
        y1: float,
        title: str | None = None,
    ) -> (CircleSVGShape, RectangleSVGShape, SquareSVGShape):
        """Make a SVG shape `plot item`

        Args:
            shape: shape type ("circle", "rectangle", "square")
            fname_or_data: filename or data
            x0: shape x0 coordinate
            y0: shape y0 coordinate
            x1: shape x1 coordinate
            y1: shape y1 coordinate
            title: label name. Default is None

        Returns:
            SVG shape
        """
        assert shape in ("circle", "rectangle", "square")
        assert isinstance(fname_or_data, (str, bytes))
        if isinstance(fname_or_data, str):
            with open(fname_or_data, "rb") as file:
                data = file.read()
        else:
            data = fname_or_data
        shapeklass = {
            "circle": CircleSVGShape,
            "rectangle": RectangleSVGShape,
            "square": SquareSVGShape,
        }[shape]
        shape = self.__shape(shapeklass, x0, y0, x1, y1, title)
        shape.set_data(data)
        return shape



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Testing plot with synchronized axes"""

from __future__ import annotations

import numpy as np
from guidata.env import execenv
from guidata.qthelpers import exec_dialog, qt_app_context
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW

from plotpy.builder import make
from plotpy.plot import BasePlot, PlotOptions
from plotpy.plot.plotwidget import SyncPlotDialog, SyncPlotWindow
from plotpy.tests.data import gen_2d_gaussian


def show_with(cls: type[SyncPlotDialog] | type[SyncPlotWindow], plot_type, *itemlists):
    """Show plot items in SyncPlotWindow or SyncPlotDialog"""
    widget = cls(
        title=f"{cls.__name__}: showing plots, optionally synchronized ({plot_type})",
        options=PlotOptions(type=plot_type),
    )
    row, col = 0, 0
    for items in itemlists:
        plot = BasePlot()
        for item in items:
            plot.add_item(item)
        plot.set_axis_font("left", QG.QFont("Courier"))
        plot.set_items_readonly(False)
        widget.add_plot(row, col, plot, sync=True)
        col += 1
        if col == 2:
            row += 1
            col = 0
    widget.finalize_configuration()
    if plot_type == "image":
        widget.get_manager().get_contrast_panel().show()
    widget.resize(800, 600)
    if cls is SyncPlotWindow:
        widget.show()
        if not execenv.unattended:
            app = QW.QApplication.instance()
            app.exec()
    else:
        exec_dialog(widget)


def test_syncplot_curves():
    """Test plot synchronization: curves"""
    x = np.linspace(-10, 10, 200)
    dy = x / 100.0
    y = np.sin(np.sin(np.sin(x)))
    x2 = np.linspace(-10, 10, 20)
    y2 = np.sin(np.sin(np.sin(x2)))
    with qt_app_context():
        itemlists = [
            [
                make.curve(x, y, color="b"),
                make.label(
                    "Relative position <b>outside</b>", (x[0], y[0]), (-10, -10), "BR"
                ),
            ],
            [make.curve(x2, y2, color="g")],
            [
                make.curve(x, np.sin(2 * y), color="r"),
                make.label(
                    "Relative position <i>inside</i>", (x[0], y[0]), (10, 10), "TL"
                ),
            ],
            [
                make.merror(x, y / 2, dy),
                make.label("Absolute position", "R", (0, 0), "R"),
                make.legend("TR"),
            ],
        ]
        for cls in (SyncPlotWindow, SyncPlotDialog):
            show_with(cls, "curve", *itemlists)


def test_syncplot_images():
    """Test plot synchronization: images"""
    img1 = gen_2d_gaussian(20, np.uint8, x0=-10.0, y0=-10.0, mu=7.0, sigma=10.0)
    img2 = gen_2d_gaussian(20, np.uint8, x0=-10.0, y0=-10.0, mu=5.0, sigma=8.0)
    img3 = gen_2d_gaussian(20, np.uint8, x0=-10.0, y0=-10.0, mu=3.0, sigma=6.0)
    with qt_app_context():

        def makeim(data):
            """Make image item"""
            return make.image(data, interpolation="nearest")

        itemlists = [[makeim(img1)], [makeim(img2)], [makeim(img3)]]

        for cls in (SyncPlotWindow, SyncPlotDialog):
            show_with(cls, "image", *itemlists)


if __name__ == "__main__":
    test_syncplot_curves()
    test_syncplot_images()



# -*- coding: utf-8 -*-

# This test was copied from ``guiqwt`` project:

# Licensed under the terms of the MIT License (see below)
#
# Copyright © 2022 Pierre Raybaut
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Dynamic curve widget test
"""

# guitest: show

import time
from datetime import datetime

import numpy as np
from guidata.qthelpers import (
    create_action,
    get_std_icon,
    win32_fix_title_bar_background,
)
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW
from qwt import QwtScaleDraw, QwtText

from plotpy.builder import make


class DummyDevice(QC.QObject):
    """Dummy device emitting random data"""

    NEW_SAMPLE = QC.Signal(object, int, float)

    def __init__(self, period=500, duration_range=None, mode=None):
        super(DummyDevice, self).__init__()
        if duration_range is None:
            duration_range = (1500, 3000)
        self.mode = "random" if mode is None else mode
        self.previous_value = None
        self.increasing_value = True
        self.duration_range = duration_range
        self.timer = QC.QTimer()
        self.timer.setInterval(period)
        self.timer.timeout.connect(self.send_data)

    def singleshot(self, slot):
        tmin, tmax = self.duration_range
        QC.QTimer.singleShot(np.random.randint(tmin, tmax), slot)

    def init_device(self, buffer_obj):
        self.NEW_SAMPLE.connect(buffer_obj.add_sample)
        self.stop_device()

    def start_device(self):
        self.timer.start()
        self.singleshot(self.stop_device)

    def stop_device(self):
        self.timer.stop()
        self.singleshot(self.start_device)

    def send_data(self):
        if self.mode == "random":
            value = np.random.rand(1)[0]
        elif self.mode == "linear":
            if self.previous_value is None:
                value = 0.0
            else:
                delta = 0.025 * (1 if self.increasing_value else -1)
                value = self.previous_value + delta
                if value <= 0.0 or value >= 1:
                    self.increasing_value = not self.increasing_value
            self.previous_value = value
        elif self.mode == "sine":
            value = 0.5 * np.sin(time.time() / np.pi) + 0.5
        else:
            raise ValueError("Unknown mode %r" % self.mode)
        quality = 0
        date = time.time()
        self.NEW_SAMPLE.emit(value, quality, date)


class CircularBuffer(QC.QObject):
    """Circular buffer storing data"""

    DATA_CHANGED = QC.Signal()

    def __init__(
        self, maxsize, period=None, dtype=None, default_value=None, maintain_value=True
    ):
        super(CircularBuffer, self).__init__()
        self._maintain_value = maintain_value
        dtype = np.float64 if dtype is None else dtype
        period = 100 if period is None else period
        self._last_sample = None
        self._previous_date = None
        self._time = np.zeros((int(maxsize),), dtype=np.float64)
        self._data = np.zeros((int(maxsize),), dtype=dtype)
        self._size = None
        self._index = None
        self._default = 0.0 if default_value is None else default_value
        self.timer = QC.QTimer()
        self.timer.setInterval(period)
        self.timer.timeout.connect(self.update_buffer)

    def init_buffer(self):
        self._size = 0
        self._index = 0
        self.timer.start()

    def update_buffer(self):
        if self._last_sample is None:
            self._last_sample = value, quality, date = self._default, -1, time.time()
            data_index = 0
            replace_sample = False
        else:
            value, quality, date = self._last_sample
            replace_sample = not self._maintain_value and self._previous_date == date
        if not replace_sample:
            if self._size == self._data.size:
                self._index = (self._index + 1) % self._data.size
            else:
                self._size += 1
        data_index = (self._index + self._size - 1) % self._data.size
        if self._maintain_value:
            time_data = time.time()
        else:
            time_data = date
        self._previous_date = date
        self._time[data_index] = time_data
        self._data[data_index] = value if quality == 0 else self._default
        self.DATA_CHANGED.emit()

    def add_sample(self, value, quality, date):
        self._last_sample = value, quality, date

    def get_data(self, nsamples=None):
        if self._size == 0:
            return np.array([]), np.array([])
        else:
            slice_range = np.arange(self._index, self._index + self._size)
            if nsamples is not None:
                slice_range = slice_range[:: slice_range.size // nsamples]
            slice_range = slice_range % self._data.size
            return (self._time[slice_range], self._data[slice_range])


class TimeScaleDraw(QwtScaleDraw):
    """Time scale draw"""

    def __init__(self):
        super(TimeScaleDraw, self).__init__()
        self.setLabelRotation(-45)
        self.setLabelAlignment(QC.Qt.AlignHCenter | QC.Qt.AlignVCenter)
        self.setSpacing(20)

    def label(self, timestamp):
        return QwtText(datetime.fromtimestamp(timestamp).strftime("%H:%M:%S"))


class DynCurveWidget(QW.QWidget):
    """Dynamic curve widget"""

    def __init__(
        self,
        title=None,
        parent=None,
        panels=None,
        period=100,
    ):
        title = self.__class__.__name__ if title is None else title
        super().__init__(parent=parent)
        self.setWindowTitle("Dynamic curve test")
        win32_fix_title_bar_background(self)

        self.plotwidget = make.widget(
            parent=self,
            toolbar=True,
            title=title,
            type="curve",
            panels=panels,
        )
        self.plotwidget.plot.set_antialiasing(True)

        self.nav_actions = []
        self.play_pause_act = None
        navtoolbar = self.create_navtoolbar()

        layout = QW.QGridLayout()
        layout.addWidget(self.plotwidget.toolbar, 0, 0, 1, 1)
        layout.addWidget(navtoolbar, 0, 1, 1, 1)
        layout.addWidget(self.plotwidget, 1, 0, 1, 2)
        self.setLayout(layout)

        self.__curve_items = {}
        self.__buffer_objs = {}
        self.__fixed_scales = False
        self.__x_origin = None
        self.__x_delta = None
        self.__auto_y_range = None
        self.set_auto_y_range(True)
        self.paused = True
        self.play_pause()
        self.initialize_plot()

    def get_plot(self):
        return self.plotwidget.get_plot()

    def initialize_plot(self):
        sdraw = TimeScaleDraw()
        plot = self.get_plot()
        plot.add_item(make.legend("TL"))
        plot.setAxisScaleDraw(plot.xBottom, sdraw)
        gridparam = plot.grid.gridparam
        gridparam.min_xenabled = gridparam.min_yenabled = False
        gridparam.background = "#646464"
        gridparam.update_grid(plot.grid)

    def create_navtoolbar(self):
        navtoolbar = QW.QToolBar("Navigation")
        for icon, tool, meth in (
            ("MediaSkipBackward", "Début", self.goto_start),
            ("MediaSeekBackward", "Arrière", lambda: self.goto(-1)),
            ("MediaPlay", "Lecture/Pause", self.play_pause),
            ("MediaSeekForward", "Avant", lambda: self.goto(1)),
            ("MediaSkipForward", "Fin", self.goto_end),
        ):
            action = create_action(self, tool, meth, icon=get_std_icon(icon))
            navtoolbar.addAction(action)
            if tool == "Lecture/Pause":
                self.play_pause_act = action
            else:
                self.nav_actions.append(action)
        self.plotwidget.manager.add_toolbar(navtoolbar, toolbar_id="navtoolbar")
        return navtoolbar

    def goto_start(self):
        self.set_time_range(origin=0.0)
        self.update_all_curves()

    def goto_end(self):
        self.set_time_range(origin=None)
        self.update_all_curves()

    def goto(self, direction):
        xmin, xmax = self.get_plot().get_axis_limits("bottom")
        origin = xmin + 0.25 * self.__x_delta * direction
        self.set_time_range(origin=origin)
        self.update_all_curves()

    def play_pause(self):
        self.paused = not self.paused
        icon_name = "MediaPlay" if self.paused else "MediaPause"
        self.play_pause_act.setIcon(get_std_icon(icon_name))
        for qobj in self.nav_actions + [self.plotwidget.toolbar]:
            qobj.setEnabled(self.paused)
        if not self.paused:
            self.set_time_range(origin=None)

    def add_curve(
        self,
        buffer_obj,
        xaxis="bottom",
        yaxis="left",
        ylimits=None,
        unit=None,
        title="",
        color=None,
        linestyle=None,
        linewidth=3.0,
        marker=None,
        markersize=None,
        markerfacecolor=None,
        markeredgecolor=None,
        shade=None,
        curvestyle=None,
    ):
        item = make.curve(
            [],
            [],
            title=title,
            xaxis=xaxis,
            yaxis=yaxis,
            color=color,
            linestyle=linestyle,
            linewidth=linewidth,
            marker=marker,
            markersize=markersize,
            markerfacecolor=markerfacecolor,
            markeredgecolor=markeredgecolor,
            shade=shade,
            curvestyle=curvestyle,
        )
        self.__curve_items[id(item)] = item
        plot = self.get_plot()
        if unit is not None:
            plot.set_axis_unit(yaxis, unit)
        if ylimits is not None:
            ymin, ymax = ylimits
            plot.set_axis_limits(yaxis, ymin, ymax)
            self.set_auto_y_range(False)
        plot.add_item(item)
        plot.enable_used_axes()
        self.__buffer_objs[id(item)] = buffer_obj
        buffer_obj.DATA_CHANGED.connect(
            lambda item_id=id(item): self.update_curve(item_id)
        )

    def set_time_range(self, origin=None, delta=None):
        if origin is not None and not isinstance(origin, float):
            raise ValueError("Invalid time origin, expected None or float")
        if delta is not None and not isinstance(delta, float):
            raise ValueError("Invalid time delta, expected float")
        self.__x_origin = origin
        if delta is not None:
            self.__x_delta = delta
        self.set_autoscale(False)

    def set_auto_y_range(self, state):
        self.__auto_y_range = state

    def set_axis_unit(self, axis_id, unit):
        self.get_plot().set_axis_unit(axis_id, unit)

    def set_autoscale(self, state):
        self.__fixed_scales = not state
        plot = self.get_plot()
        if state:
            plot.do_autoscale()
        # else:
        #     plot.disable_autoscale()

    def update_scale(self, item, xdata, ydata):
        plot = self.get_plot()
        if self.__fixed_scales:
            if self.__x_origin is None:
                # Xmin is set automatically to adjust Xrange to [now-delta;now]
                now = time.time()
                xmin, xmax = now - self.__x_delta, now
            else:
                min_xdata = xdata.min()
                xmin = max([self.__x_origin, min_xdata])
                xmax = xmin + self.__x_delta
            plot.set_axis_limits(item.xAxis(), xmin, xmax)
            if self.__auto_y_range:
                ydata_min, ydata_max = ydata.min(), ydata.max()
                delta_y = 0.1 * (ydata_max - ydata_min)
                ymin, ymax = ydata_min - delta_y, ydata_max + delta_y
                plot.set_axis_limits(item.yAxis(), ymin, ymax)
        plot.replot()

    def update_curve(self, item_id, force=False):
        if self.paused and not force:
            return
        xdata, ydata = self.__buffer_objs[item_id].get_data()
        item = self.__curve_items[item_id]
        item.set_data(xdata, ydata)
        self.update_scale(item, xdata, ydata)

    def update_all_curves(self, force=True):
        for item_id in self.__curve_items:
            self.update_curve(item_id, force=force)


if __name__ == "__main__":
    from guidata import qapplication

    app = qapplication()

    buffer1 = CircularBuffer(1e6, period=100)
    buffer1.init_buffer()
    buffer2 = CircularBuffer(1e6, period=100)
    buffer2.init_buffer()

    win = DynCurveWidget()
    win.set_time_range(delta=10.0)
    win.set_auto_y_range(True)
    win.add_curve(
        buffer1,
        title="Temperature",
        yaxis="left",
        ylimits=(0.0, 1.0),
        unit="°C",
        color="b",
    )
    win.add_curve(
        buffer2,
        title="Pressure",
        yaxis="right",
        ylimits=(0.0, 1.0),
        unit="kPa",
        color="r",
    )

    device1 = DummyDevice(period=10, mode="sine")
    device1.init_device(buffer1)
    device2 = DummyDevice(period=500, mode="sine")  # , duration_range=(4000, 7000))
    device2.init_device(buffer2)
    #    win.showMaximized()
    win.resize(800, 400)
    win.show()
    app.exec()
